from __future__ import annotations

import logging
import threading
import time
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, List

import numpy as np
import numpy.typing as npt
import pyarrow as pa
import requests
from astroquery.jplsbdb import SBDB

from ...coordinates.cometary import CometaryCoordinates
from ...coordinates.covariances import CoordinateCovariances, sigmas_to_covariances
from ...coordinates.origin import Origin
from ...time import Timestamp
from ..orbits import Orbits

logger = logging.getLogger(__name__)

_SBDB_API_URL = "https://ssd-api.jpl.nasa.gov/sbdb.api"
_SBDB_API_FAIR_USE_MAX_CONCURRENT_REQUESTS = 1

_thread_local = threading.local()


def _get_requests_session() -> requests.Session:
    """
    Return a per-thread `requests.Session`.

    Why: using a session enables connection pooling, which reduces overhead when querying many IDs.
    A per-thread session is a safe default if callers choose to enable limited concurrency.
    """
    sess = getattr(_thread_local, "session", None)
    if sess is None:
        sess = requests.Session()
        _thread_local.session = sess
    return sess


def _convert_SBDB_covariances(
    sbdb_covariances: npt.ArrayLike,
) -> npt.ArrayLike:
    """
    Convert SBDB covariance matrices to Cometary covariance matrices.

    Parameters
    ----------
    sbdb_covariances : `~numpy.ndarray` (N, 6, 6)
        Covariance matrices pulled from JPL's Small Body Database Browser.

    Returns
    -------
    covariances : `~numpy.ndarray` (N, 6, 6)
        Cometary covariance matrices.
    """
    covariances = np.zeros_like(sbdb_covariances)
    # sigma_q{x}
    covariances[:, 0, 0] = sbdb_covariances[:, 1, 1]  # sigma_qq
    covariances[:, 1, 0] = covariances[:, 0, 1] = sbdb_covariances[:, 0, 1]  # sigma_qe
    covariances[:, 2, 0] = covariances[:, 0, 2] = sbdb_covariances[:, 5, 1]  # sigma_qi
    covariances[:, 3, 0] = covariances[:, 0, 3] = sbdb_covariances[
        :, 3, 1
    ]  # sigma_qraan
    covariances[:, 4, 0] = covariances[:, 0, 4] = sbdb_covariances[:, 4, 1]  # sigma_qap
    covariances[:, 5, 0] = covariances[:, 0, 5] = sbdb_covariances[:, 2, 1]  # sigma_qtp

    # sigma_e{x}
    covariances[:, 1, 1] = sbdb_covariances[:, 0, 0]  # sigma_ee
    covariances[:, 2, 1] = covariances[:, 1, 2] = sbdb_covariances[:, 5, 0]  # sigma_ei
    covariances[:, 3, 1] = covariances[:, 1, 3] = sbdb_covariances[
        :, 3, 0
    ]  # sigma_eraan
    covariances[:, 4, 1] = covariances[:, 1, 4] = sbdb_covariances[:, 4, 0]  # sigma_eap
    covariances[:, 5, 1] = covariances[:, 1, 5] = sbdb_covariances[:, 2, 0]  # sigma_etp

    # sigma_i{x}
    covariances[:, 2, 2] = sbdb_covariances[:, 5, 5]  # sigma_ii
    covariances[:, 3, 2] = covariances[:, 2, 3] = sbdb_covariances[
        :, 3, 5
    ]  # sigma_iraan
    covariances[:, 4, 2] = covariances[:, 2, 4] = sbdb_covariances[:, 4, 5]  # sigma_iap
    covariances[:, 5, 2] = covariances[:, 2, 5] = sbdb_covariances[:, 2, 5]  # sigma_itp

    # sigma_raan{x}
    covariances[:, 3, 3] = sbdb_covariances[:, 3, 3]  # sigma_raanraan
    covariances[:, 4, 3] = covariances[:, 3, 4] = sbdb_covariances[
        :, 4, 3
    ]  # sigma_raanap
    covariances[:, 5, 3] = covariances[:, 3, 5] = sbdb_covariances[
        :, 2, 3
    ]  # sigma_raantp

    # sigma_ap{x}
    covariances[:, 4, 4] = sbdb_covariances[:, 4, 4]  # sigma_apap
    covariances[:, 5, 4] = covariances[:, 4, 5] = sbdb_covariances[
        :, 2, 4
    ]  # sigma_aptp

    # sigma_tp{x}
    covariances[:, 5, 5] = sbdb_covariances[:, 2, 2]  # sigma_tptp

    return covariances


def _get_sbdb_elements(obj_ids: List[str]) -> List[OrderedDict]:
    """
    Get orbital elements and other object properties
    from JPL's Small Body Database Browser.

    Parameters
    ----------
    obj_ids : List
        Object IDs to query.

    Returns
    -------
    results : List
        List of dictionaries containing orbital elements and other object properties.
    """
    results = []
    SBDB.clear_cache()  # Yikes!
    for obj_id in obj_ids:
        result = SBDB.query(
            obj_id,
            covariance="mat",
            id_type="search",
            full_precision=True,
            solution_epoch=False,
        )
        results.append(result)

    return results


def _orbits_from_sbdb_results(ids: npt.ArrayLike, results: List[OrderedDict]) -> Orbits:
    """
    Convert SBDB query results into an `Orbits` table.

    What: shared implementation for both the legacy astroquery-based query and the new
    direct-HTTP query.
    Why: keeping a single conversion path ensures both entrypoints return identical results.
    """
    orbit_ids = []
    object_ids = []
    classes = []
    coords_cometary = np.zeros((len(results), 6), dtype=np.float64)
    covariances_sbdb = np.zeros((len(results), 6, 6), dtype=np.float64)
    times = np.zeros((len(results)), dtype=np.float64)

    for i, result in enumerate(results):
        if "object" not in result:
            raise NotFoundError("object {} was not found", ids[i])

        orbit_ids.append(f"{i:05d}")
        object_ids.append(result["object"]["fullname"])
        classes.append(result["object"]["orbit_class"]["code"])

        orbit = result["orbit"]
        elements = orbit["elements"]
        epoch = orbit["epoch"]
        if "covariance" in orbit:
            labels = orbit["covariance"]["labels"]
            if len(labels) > 6:
                logger.debug(
                    "Covariance matrix has more parameters than just orbital elements. "
                    "Ignoring non-orbital elements in covariance matrix."
                )
                labels = labels[:6]

            expected_labels = ["e", "q", "tp", "node", "peri", "i"]
            if labels != expected_labels:
                raise ValueError(
                    "Expected covariance matrix labels to be {expected_labels} instead got {labels}."
                )

            # Limit covariances to just the orbital elements
            # The SBDB API documentation states that physical parameter covariances
            # are appended to the rows and columns of the covariance matrix with the
            # orbital elements remaining in the first 6 rows and columns.
            # See: Orbit Subsection: covariance in https://ssd-api.jpl.nasa.gov/doc/sbdb.html
            covariances_sbdb[i, :, :] = orbit["covariance"]["data"][:6, :6]

            if "elements" in orbit["covariance"]:
                # If elements is provided inside covariance, then it's
                # the elements at the epoch which was used to
                # calculate covariance, so we should prefer it.
                elements = orbit["covariance"]["elements"]
                epoch = orbit["covariance"]["epoch"]

        else:
            sigmas = np.array(
                [
                    [
                        elements["e_sig"],
                        elements["q_sig"].value,
                        elements["tp_sig"].value,
                        elements["om_sig"].value,
                        elements["w_sig"].value,
                        elements["i_sig"].value,
                    ]
                ]
            )
            covariances_sbdb[i, :, :] = sigmas_to_covariances(sigmas)[0]

        times[i] = epoch.value
        coords_cometary[i, 0] = elements["q"].value
        coords_cometary[i, 1] = elements["e"]
        coords_cometary[i, 2] = elements["i"].value
        coords_cometary[i, 3] = elements["om"].value
        coords_cometary[i, 4] = elements["w"].value
        coords_cometary[i, 5] = (
            Timestamp.from_jd([elements["tp"].value], scale="tdb").mjd()[0].as_py()
        )

    covariances_cometary = _convert_SBDB_covariances(covariances_sbdb)
    times = Timestamp.from_jd(times, scale="tdb")
    origin = Origin.from_kwargs(code=["SUN" for i in range(len(times))])
    frame = "ecliptic"
    coordinates = CometaryCoordinates.from_kwargs(
        time=times,
        q=coords_cometary[:, 0],
        e=coords_cometary[:, 1],
        i=coords_cometary[:, 2],
        raan=coords_cometary[:, 3],
        ap=coords_cometary[:, 4],
        tp=coords_cometary[:, 5],
        covariance=CoordinateCovariances.from_matrix(covariances_cometary),
        origin=origin,
        frame=frame,
    )

    orbit_ids = np.array(orbit_ids, dtype="object")
    object_ids = np.array(object_ids, dtype="object")
    classes = np.array(classes)

    return Orbits.from_kwargs(
        orbit_id=orbit_ids, object_id=object_ids, coordinates=coordinates.to_cartesian()
    )


def query_sbdb(ids: npt.ArrayLike) -> Orbits:
    """
    Query JPL's Small-Body Database (SBDB) for orbits. The epoch at
    which the orbits are returned are near the epoch as published by the
    Minor Planet Center.

    By default, the orbit's covariance matrices are also queried for. If they
    are not available, then the 1-sigma uncertainties are used to construct
    the covariance matrices.

    Parameters
    ----------
    ids : list
        List of object IDs to query.

    Returns
    -------
    orbits : `~adam_core.orbits.orbits.Orbits`
        Orbits object containing the queried orbits.

    Raises
    ------
    NotFoundError: If any of the queries object IDs are not found.
    """
    results = _get_sbdb_elements(ids)
    return _orbits_from_sbdb_results(ids, results)


def _sbdb_api_get_json(
    object_id: str,
    *,
    timeout_s: float,
    max_attempts: int,
) -> dict[str, Any]:
    """
    Query JPL's public SBDB JSON API for a single object, with retries.

    Why: we want explicit timeout/retry behavior, and we want to avoid global cache clearing.

    Notes:
    - Per the JPL SSD/CNEOS API fair use policy, clients should not send concurrent requests.
      This helper does not enforce that policy; the public entrypoint defaults to sequential
      requests (max_concurrent_requests=1).
    """
    obj = str(object_id).strip()
    if not obj:
        raise ValueError("object_id must be non-empty")
    if timeout_s <= 0:
        raise ValueError("timeout_s must be > 0")
    if max_attempts <= 0:
        raise ValueError("max_attempts must be > 0")

    params = {
        "sstr": obj,
        "cov": "mat",
        "full-prec": "true",
    }

    last_err: Exception | None = None
    for attempt in range(max_attempts):
        try:
            resp = _get_requests_session().get(
                _SBDB_API_URL, params=params, timeout=timeout_s
            )
            resp.raise_for_status()
            return resp.json()
        except (
            requests.exceptions.Timeout,
            requests.exceptions.ConnectionError,
        ) as err:
            last_err = err
        except requests.exceptions.HTTPError as err:
            # Retry on transient server errors and explicit throttling.
            status = err.response.status_code if err.response is not None else None
            if status is not None and (status >= 500 or status == 429):
                last_err = err
            else:
                raise
        except Exception:
            # Non-retryable (JSON decode, unexpected failure, etc.)
            raise

        # Exponential backoff with a small cap.
        sleep_s = min(8.0, 0.5 * (2**attempt))
        time.sleep(sleep_s)

    raise RuntimeError(f"SBDB query failed after {max_attempts} attempts: {last_err}")


def _sbdb_float(value: Any) -> float:
    """
    Convert a JSON scalar (string/number) into a float.

    SBDB returns many numeric fields as strings (including scientific notation). We normalize
    those to floats for orbit construction.
    """
    if value is None:
        raise ValueError("Expected a numeric value, got None.")
    if isinstance(value, (float, int, np.floating, np.integer)):
        return float(value)
    if isinstance(value, str):
        return float(value)
    raise TypeError(f"Expected a numeric JSON scalar, got {type(value)!r}.")


def _sbdb_elements_map(elements: Any) -> dict[str, dict[str, Any]]:
    """
    Convert SBDB's `orbit.elements` list into a dict keyed by element short-name.
    """
    if not isinstance(elements, list):
        raise ValueError("Expected SBDB orbit elements to be a list.")
    out: dict[str, dict[str, Any]] = {}
    for el in elements:
        if not isinstance(el, dict):
            continue
        name = el.get("name")
        if name is None:
            continue
        out[str(name)] = el
    return out


def _sbdb_element_value(
    elements_by_name: dict[str, dict[str, Any]], name: str
) -> float:
    el = elements_by_name.get(name)
    if el is None:
        raise ValueError(f"SBDB orbit elements missing {name!r}.")
    if "value" not in el or el["value"] is None:
        raise ValueError(f"SBDB orbit element {name!r} is missing a value.")
    return _sbdb_float(el["value"])


def _sbdb_element_sigma(
    elements_by_name: dict[str, dict[str, Any]], name: str
) -> float:
    el = elements_by_name.get(name)
    if el is None:
        raise ValueError(f"SBDB orbit elements missing {name!r}.")
    sigma = el.get("sigma")
    if sigma is None:
        # Some SBDB payloads omit per-element uncertainties when no covariance matrix is provided.
        # We treat this as "unknown uncertainty" and propagate NaNs through the diagonal fallback.
        return float("nan")
    try:
        return _sbdb_float(sigma)
    except Exception:
        return float("nan")


def _orbits_from_sbdb_payloads(
    ids: list[str],
    payloads: list[dict[str, Any]],
) -> Orbits:
    """
    Convert raw SBDB JSON payloads into an `Orbits` table.

    This mirrors the behavior of the legacy `query_sbdb` implementation:
    - Prefer covariance-provided elements/epoch when present.
    - Use covariance matrix when available; otherwise build a diagonal covariance from sigmas.
    """
    if len(ids) != len(payloads):
        raise ValueError("ids and payloads must have the same length.")

    expected_labels = ["e", "q", "tp", "node", "peri", "i"]

    orbit_ids: list[str] = []
    object_ids: list[str] = []

    coords_cometary = np.zeros((len(payloads), 6), dtype=np.float64)
    covariances_sbdb = np.zeros((len(payloads), 6, 6), dtype=np.float64)
    times_jd = np.zeros((len(payloads)), dtype=np.float64)

    for i, (obj_id, payload) in enumerate(zip(ids, payloads)):
        if "object" not in payload:
            raise NotFoundError("object {} was not found", obj_id)
        if "orbit" not in payload:
            raise ValueError(f"SBDB payload for {obj_id!r} missing 'orbit'.")

        obj = payload["object"] or {}
        orbit_ids.append(f"{i:05d}")
        object_ids.append(str(obj.get("fullname")))

        orbit = payload["orbit"] or {}
        elements_list = orbit.get("elements")
        epoch_jd = _sbdb_float(orbit.get("epoch"))

        cov = orbit.get("covariance")
        cov_matrix: np.ndarray | None = None
        if isinstance(cov, dict) and cov.get("data") is not None:
            labels = cov.get("labels")
            if isinstance(labels, list):
                labels6 = [str(x) for x in labels[:6]]
                if labels6 != expected_labels:
                    raise ValueError(
                        f"Expected covariance matrix labels to be {expected_labels} "
                        f"in the first 6 entries, got {labels6}."
                    )

            data = np.asarray(cov["data"], dtype=np.float64)
            if data.ndim != 2 or data.shape[0] < 6 or data.shape[1] < 6:
                raise ValueError("Expected SBDB covariance matrix to be at least 6x6.")
            cov_matrix = data[:6, :6]

            # If covariance provides elements, prefer them (and the covariance epoch).
            if "elements" in cov and cov["elements"] is not None:
                elements_list = cov["elements"]
                if cov.get("epoch") is not None:
                    epoch_jd = _sbdb_float(cov.get("epoch"))

        if elements_list is None:
            raise ValueError(f"SBDB payload for {obj_id!r} missing orbit elements.")

        elements_by_name = _sbdb_elements_map(elements_list)

        if cov_matrix is None:
            # Fallback: build a diagonal covariance from per-element sigmas.
            sigmas = np.array(
                [
                    [
                        _sbdb_element_sigma(elements_by_name, "e"),
                        _sbdb_element_sigma(elements_by_name, "q"),
                        _sbdb_element_sigma(elements_by_name, "tp"),
                        _sbdb_element_sigma(elements_by_name, "om"),
                        _sbdb_element_sigma(elements_by_name, "w"),
                        _sbdb_element_sigma(elements_by_name, "i"),
                    ]
                ],
                dtype=np.float64,
            )
            cov_matrix = sigmas_to_covariances(sigmas)[0]

        covariances_sbdb[i, :, :] = cov_matrix

        times_jd[i] = epoch_jd

        q = _sbdb_element_value(elements_by_name, "q")
        e = _sbdb_element_value(elements_by_name, "e")
        inc = _sbdb_element_value(elements_by_name, "i")
        om = _sbdb_element_value(elements_by_name, "om")
        w = _sbdb_element_value(elements_by_name, "w")
        tp_jd = _sbdb_element_value(elements_by_name, "tp")
        tp_mjd = Timestamp.from_jd([tp_jd], scale="tdb").mjd()[0].as_py()

        coords_cometary[i, 0] = q
        coords_cometary[i, 1] = e
        coords_cometary[i, 2] = inc
        coords_cometary[i, 3] = om
        coords_cometary[i, 4] = w
        coords_cometary[i, 5] = tp_mjd

    covariances_cometary = _convert_SBDB_covariances(covariances_sbdb)
    times = Timestamp.from_jd(times_jd, scale="tdb")
    origin = Origin.from_kwargs(code=["SUN" for _ in range(len(times))])

    coordinates = CometaryCoordinates.from_kwargs(
        time=times,
        q=coords_cometary[:, 0],
        e=coords_cometary[:, 1],
        i=coords_cometary[:, 2],
        raan=coords_cometary[:, 3],
        ap=coords_cometary[:, 4],
        tp=coords_cometary[:, 5],
        covariance=CoordinateCovariances.from_matrix(covariances_cometary),
        origin=origin,
        frame="ecliptic",
    )

    return Orbits.from_kwargs(
        orbit_id=np.array(orbit_ids, dtype="object"),
        object_id=np.array(object_ids, dtype="object"),
        coordinates=coordinates.to_cartesian(),
    )


def _get_sbdb_payloads_new(
    obj_ids: List[str],
    *,
    max_concurrent_requests: int,
    timeout_s: float,
    max_attempts: int,
) -> list[dict[str, Any]]:
    """
    Fetch SBDB JSON payloads via direct HTTP, optionally with limited concurrency.

    Important: JPL's SSD/CNEOS API fair use policy requests only one in-flight request at a time.
    The public entrypoint defaults to `max_concurrent_requests=1` to comply.
    """
    if max_concurrent_requests <= 0:
        raise ValueError("max_concurrent_requests must be > 0")

    n = len(obj_ids)
    if n == 0:
        return []

    max_workers = min(int(max_concurrent_requests), n)
    if max_workers > 1:
        logger.warning(
            "query_sbdb_new is configured with max_concurrent_requests=%s. "
            "JPL's SSD/CNEOS API fair use policy requests only one in-flight request at a time; "
            "concurrent requests may be rejected.",
            max_workers,
        )

    results: list[dict[str, Any] | None] = [None] * n

    def fetch_one(i: int, object_id: str) -> tuple[int, dict[str, Any]]:
        payload = _sbdb_api_get_json(
            object_id, timeout_s=timeout_s, max_attempts=max_attempts
        )
        return i, payload

    with ThreadPoolExecutor(max_workers=max_workers) as ex:
        futures = [ex.submit(fetch_one, i, obj_id) for i, obj_id in enumerate(obj_ids)]
        for fut in as_completed(futures):
            i, res = fut.result()
            results[i] = res

    out: list[dict[str, Any]] = []
    for r in results:
        if r is None:
            raise RuntimeError("SBDB payload missing from concurrent fetch.")
        out.append(r)
    return out


def query_sbdb_new(
    ids: npt.ArrayLike,
    *,
    max_concurrent_requests: int = 1,
    timeout_s: float = 60.0,
    max_attempts: int = 5,
    allow_missing: bool = False,
    orbit_id_from_input: bool = False,
) -> Orbits:
    """
    Query JPL SBDB for orbits using direct HTTP requests (new implementation).

    This is intended to be a drop-in alternative to `query_sbdb` that:
    - avoids `SBDB.clear_cache()`,
    - provides explicit timeout/retry controls, and
    - can optionally fetch multiple objects concurrently.

    Notes
    -----
    JPL's SSD/CNEOS API fair use policy requests only one in-flight API request at a time.
    Therefore, the default `max_concurrent_requests=1` is the recommended setting.

    Parameters
    ----------
    allow_missing : bool, optional
        If True, do not raise when an ID is not found in SBDB. Instead, return an `Orbits` table
        containing only the successfully resolved IDs (potentially empty).
    orbit_id_from_input : bool, optional
        If True, set the returned `Orbits.orbit_id` values to the input IDs (after any missing
        filtering). This is useful when callers need to map rows back to the requested identifiers.
    """
    # Normalize ids into a list of strings while preserving the caller's order.
    if isinstance(ids, (str, bytes)):
        obj_ids = [str(ids)]
    else:
        obj_ids = [str(x) for x in ids]

    payloads = _get_sbdb_payloads_new(
        obj_ids,
        max_concurrent_requests=max_concurrent_requests,
        timeout_s=timeout_s,
        max_attempts=max_attempts,
    )
    if allow_missing:
        kept_ids: list[str] = []
        kept_payloads: list[dict[str, Any]] = []
        for obj_id, payload in zip(obj_ids, payloads):
            if "object" not in payload:
                continue
            kept_ids.append(obj_id)
            kept_payloads.append(payload)

        if not kept_ids:
            return Orbits.empty()

        orbits = _orbits_from_sbdb_payloads(kept_ids, kept_payloads)
        if orbit_id_from_input:
            orbits = orbits.set_column(
                "orbit_id", pa.array(kept_ids, type=pa.large_string())
            )
        return orbits

    orbits = _orbits_from_sbdb_payloads(obj_ids, payloads)
    if orbit_id_from_input:
        orbits = orbits.set_column(
            "orbit_id", pa.array(obj_ids, type=pa.large_string())
        )
    return orbits


class NotFoundError(Exception):
    def __init__(self, message, object_id):
        self.message = message
        self.object_id = object_id

    def __str__(self):
        return self.message



from __future__ import annotations

from dataclasses import dataclass
from typing import List

import numpy as np
import pyarrow.compute as pc
import quivr as qv

from ..constants import KM_P_AU
from ..time import Timestamp
from . import cartesian
from .covariances import CoordinateCovariances, transform_covariances_jacobian
from .origin import Origin

__all__ = [
    "GeodeticCoordinates",
]


@dataclass
class GeodeticConstants:
    """
    Geodetic constants for the Earth.

    Parameters
    ----------
    a : float
        Semi-major axis of the Earth in units of distance.
    b : float
        Semi-minor axis of the Earth in units of distance.
    f : float
        Flattening of the Earth.
    """

    a: float
    b: float
    f: float


WGS84 = GeodeticConstants(
    a=6378137.0 / KM_P_AU / 1000,
    b=6356752.31424518 / KM_P_AU / 1000,
    f=1.0 / 298.257223563,
)


class GeodeticCoordinates(qv.Table):

    alt = qv.Float64Column(nullable=True)
    lon = qv.Float64Column(nullable=True)
    lat = qv.Float64Column(nullable=True)
    vup = qv.Float64Column(nullable=True)
    veast = qv.Float64Column(nullable=True)
    vnorth = qv.Float64Column(nullable=True)
    time = Timestamp.as_column(nullable=True)
    covariance = CoordinateCovariances.as_column(nullable=True)
    origin = Origin.as_column()
    frame = qv.StringAttribute(default="unspecified")

    @property
    def values(self) -> np.ndarray:
        return np.array(
            self.table.select(["alt", "lon", "lat", "vup", "veast", "vnorth"])
        )

    @property
    def sigma_alt(self):
        """
        1-sigma uncertainty in altitude.
        """
        return self.covariance.sigmas[:, 0]

    @property
    def sigma_lon(self):
        """
        1-sigma uncertainty in longitude.
        """
        return self.covariance.sigmas[:, 1]

    @property
    def sigma_lat(self):
        """
        1-sigma uncertainty in latitude.
        """
        return self.covariance.sigmas[:, 2]

    @property
    def sigma_vup(self):
        """
        1-sigma uncertainty in up velocity.
        """
        return self.covariance.sigmas[:, 3]

    @property
    def sigma_veast(self):
        """
        1-sigma uncertainty in east velocity.
        """
        return self.covariance.sigmas[:, 4]

    @property
    def sigma_vnorth(self):
        """
        1-sigma uncertainty in north velocity.
        """
        return self.covariance.sigmas[:, 5]

    def google_maps_url(self, zoom: int = 18) -> List[str]:
        """
        Generate a Google Maps URL for the coordinates.
        """
        urls = []
        for lat, lon in zip(self.lat, self.lon):
            urls.append(
                f"https://www.google.com/maps/@{lat.as_py()},{lon.as_py() - 360},{zoom}z"
            )
        return urls

    @classmethod
    def from_cartesian(
        cls, cartesian: cartesian.CartesianCoordinates
    ) -> "GeodeticCoordinates":
        from .transform import _cartesian_to_geodetic, cartesian_to_geodetic

        assert (
            cartesian.frame == "itrf93"
        ), "Cartesian coordinates must be in ITRF93 frame"
        assert pc.all(
            pc.equal(cartesian.origin.code, "EARTH")
        ), "Cartesian coordinates must be in Earth-centered frame"

        coords_geodetic = cartesian_to_geodetic(cartesian.values, a=WGS84.a, f=WGS84.f)
        coords_geodetic = np.array(coords_geodetic)

        if not cartesian.covariance.is_all_nan():
            cartesian_covariances = cartesian.covariance.to_matrix()
            covariances_geodetic = transform_covariances_jacobian(
                cartesian.values, cartesian_covariances, _cartesian_to_geodetic
            )
        else:
            covariances_geodetic = np.empty(
                (len(coords_geodetic), 6, 6), dtype=np.float64
            )
            covariances_geodetic.fill(np.nan)

        covariances_geodetic = CoordinateCovariances.from_matrix(covariances_geodetic)
        coords = cls.from_kwargs(
            alt=coords_geodetic[:, 0],
            lon=coords_geodetic[:, 1],
            lat=coords_geodetic[:, 2],
            vup=coords_geodetic[:, 3],
            veast=coords_geodetic[:, 4],
            vnorth=coords_geodetic[:, 5],
            time=cartesian.time,
            covariance=covariances_geodetic,
            origin=cartesian.origin,
            frame=cartesian.frame,
        )

        return coords



import multiprocessing as mp
from typing import Dict, List, Optional, Tuple

import jax.numpy as jnp
import numpy as np
import pyarrow as pa
import quivr as qv
import ray
from jax import jit, lax, vmap
from ray import ObjectRef

from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.covariances import (
    CoordinateCovariances,
    transform_covariances_jacobian,
)
from ..coordinates.origin import Origin, OriginCodes
from ..coordinates.spherical import SphericalCoordinates
from ..coordinates.transform import _cartesian_to_spherical, transform_coordinates
from ..observers.observers import Observers
from ..orbits.ephemeris import Ephemeris
from ..orbits.orbits import Orbits
from ..photometry.magnitude import (
    calculate_apparent_magnitude_v,
    calculate_apparent_magnitude_v_and_phase_angle,
    calculate_phase_angle,
)
from ..ray_cluster import initialize_use_ray
from ..utils.chunking import process_in_chunks
from ..utils.iter import _iterate_chunks
from .aberrations import _add_light_time, add_stellar_aberration


@jit
def _generate_ephemeris_2body(
    propagated_orbit: np.ndarray,
    observation_time: float,
    observer_coordinates: jnp.ndarray,
    mu: float,
    lt_tol: float = 1e-10,
    max_iter: int = 100,
    tol: float = 1e-15,
    stellar_aberration: bool = False,
) -> Tuple[jnp.ndarray, jnp.float64, jnp.ndarray]:
    """
    Given a propagated orbit, generate its on-sky ephemeris as viewed from the observer.
    This function calculates the light time delay between the propagated orbit and the observer,
    and then propagates the orbit backward by that amount to when the light from object was actually
    emitted towards the observer ("astrometric coordinates").

    The motion of the observer in an inertial frame will cause an object
    to appear in a different location than its true location, this is known as
    stellar aberration (often referred to in combination with other aberrations as "apparent
    coordinates"). Stellar aberration can optionally be applied after
    light time correction has been added but it should not be necessary when comparing to ephemerides
    of solar system small bodies extracted from astrometric catalogs. The stars to which the
    catalog is calibrated undergo the same aberration as the moving objects as seen from the observer.

    If stellar aberration is applied then the velocity of the input orbits are unmodified, only the position
    vector is modified with stellar aberration.

    For more details on aberrations see:
        https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/FORTRAN/req/abcorr.html
        https://ssd.jpl.nasa.gov/horizons/manual.html#defs

    Parameters
    ----------
    propagated_orbit : `~jax.numpy.ndarray` (6)
        Barycentric Cartesian orbit propagated to the given time.
    observation_time : float
        Epoch at which orbit and observer coordinates are defined.
    observer_coordinates : `~jax.numpy.ndarray` (3)
        Barycentric Cartesian observer coordinates.
    mu : float (1)
        Gravitational parameter (GM) of the attracting body in units of
        AU**3 / d**2.
    lt_tol : float, optional
        Calculate aberration to within this value in time (units of days).
    max_iter : int, optional
        Maximum number of iterations over which to converge for propagation.
    tol : float, optional
        Numerical tolerance to which to compute universal anomaly during propagation using the Newtown-Raphson
        method.
    stellar_aberration : bool, optional
        Apply stellar aberration to the ephemerides.

    Returns
    -------
    ephemeris_spherical : `~jax.numpy.ndarray` (6)
        Topocentric Spherical ephemeris.
    lt : float
        Light time correction (t0 - corrected_t0).
    aberrated_orbit : `~jax.numpy.ndarray` (6)
        Barycentric Cartesian orbit corrected for light time (emission time state).
    """
    # Add light time correction
    propagated_orbits_aberrated, light_time = _add_light_time(
        propagated_orbit,
        observation_time,
        observer_coordinates[0:3],
        lt_tol=lt_tol,
        mu=mu,
        max_iter=max_iter,
        tol=tol,
    )

    # Calculate topocentric coordinates
    topocentric_coordinates = propagated_orbits_aberrated - observer_coordinates

    # Apply stellar aberration to topocentric coordinates
    topocentric_coordinates = lax.cond(
        stellar_aberration,
        lambda topocentric_coords: topocentric_coords.at[0:3].set(
            add_stellar_aberration(
                propagated_orbits_aberrated.reshape(1, -1),
                observer_coordinates.reshape(1, -1),
            )[0],
        ),
        lambda topocentric_coords: topocentric_coords,
        topocentric_coordinates,
    )

    # Convert to spherical coordinates
    ephemeris_spherical = _cartesian_to_spherical(topocentric_coordinates)

    return ephemeris_spherical, light_time, propagated_orbits_aberrated


# Vectorization Map: _generate_ephemeris_2body
_generate_ephemeris_2body_vmap = jit(
    vmap(
        _generate_ephemeris_2body,
        in_axes=(0, 0, 0, 0, None, None, None, None),
        out_axes=(0, 0, 0),
    )
)


def _generate_ephemeris_2body_serial(
    propagated_orbits: Orbits,
    observers: Observers,
    *,
    lt_tol: float,
    max_iter: int,
    tol: float,
    stellar_aberration: bool,
    predict_magnitudes: bool,
    predict_phase_angle: bool,
) -> Ephemeris:
    # Delegate to the public function's existing implementation, but without Ray.
    return generate_ephemeris_2body(
        propagated_orbits,
        observers,
        lt_tol=lt_tol,
        max_iter=max_iter,
        tol=tol,
        stellar_aberration=stellar_aberration,
        predict_magnitudes=predict_magnitudes,
        predict_phase_angle=predict_phase_angle,
        max_processes=1,
    )


@ray.remote
def ephemeris_2body_worker_ray(
    start: int,
    idx_chunk: np.ndarray,
    propagated_orbits: Orbits,
    observers: Observers,
    lt_tol: float,
    max_iter: int,
    tol: float,
    stellar_aberration: bool,
    predict_magnitudes: bool,
    predict_phase_angle: bool,
) -> Tuple[int, Ephemeris]:
    prop_chunk = propagated_orbits.take(idx_chunk)
    obs_chunk = observers.take(idx_chunk)
    eph = _generate_ephemeris_2body_serial(
        prop_chunk,
        obs_chunk,
        lt_tol=lt_tol,
        max_iter=max_iter,
        tol=tol,
        stellar_aberration=stellar_aberration,
        predict_magnitudes=predict_magnitudes,
        predict_phase_angle=predict_phase_angle,
    )
    return start, eph


def generate_ephemeris_2body(
    propagated_orbits: Orbits,
    observers: Observers,
    lt_tol: float = 1e-10,
    max_iter: int = 1000,
    tol: float = 1e-15,
    stellar_aberration: bool = False,
    predict_magnitudes: bool = True,
    *,
    predict_phase_angle: bool = False,
    max_processes: Optional[int] = 1,
    chunk_size: int = 100,
) -> Ephemeris:
    """
    Generate on-sky ephemerides for each propagated orbit as viewed by the observers.
    This function calculates the light time delay between the propagated orbit and the observer,
    and then propagates the orbit backward by that amount to when the light from object was actually
    emitted towards the observer ("astrometric coordinates").

    The motion of the observer in an inertial frame will cause an object
    to appear in a different location than its true location, this is known as
    stellar aberration (often referred to in combination with other aberrations as "apparent
    coordinates"). Stellar aberration can optionally be applied after
    light time correction has been added but it should not be necessary when comparing to ephemerides
    of solar system small bodies extracted from astrometric catalogs. The stars to which the
    catalog is calibrated undergo the same aberration as the moving objects as seen from the observer.

    If stellar aberration is applied then the velocity of the input orbits are unmodified, only the position
    vector is modified with stellar aberration.

    For more details on aberrations see:
        https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/FORTRAN/req/abcorr.html
        https://ssd.jpl.nasa.gov/horizons/manual.html#defs


    Parameters
    ----------
    propagated_orbits : `~adam_core.orbits.orbits.Orbits` (N)
        Propagated orbits.
    observers : `~adam_core.observers.observers.Observers` (N)
        Observers for which to generate ephemerides. Orbits should already have been
        propagated to the same times as the observers.
    lt_tol : float, optional
        Calculate aberration to within this value in time (units of days).
    mu : float, optional
        Gravitational parameter (GM) of the attracting body in units of
        AU**3 / d**2.
    max_iter : int, optional
        Maximum number of iterations over which to converge for propagation.
    tol : float, optional
        Numerical tolerance to which to compute universal anomaly during propagation using the Newtown-Raphson
        method.
    stellar_aberration : bool, optional
        Apply stellar aberration to the ephemerides.

    Returns
    -------
    ephemeris : `~adam_core.orbits.ephemeris.Ephemeris` (N)
        Topocentric ephemerides for each propagated orbit as observed by the given observers.
    """
    if max_processes is None:
        max_processes = mp.cpu_count()

    if max_processes > 1:
        initialize_use_ray(num_cpus=max_processes)

        num_entries = len(observers)
        assert len(propagated_orbits) == num_entries

        propagated_ref = ray.put(propagated_orbits)  # type: ignore[name-defined]
        observers_ref = ray.put(observers)  # type: ignore[name-defined]

        idx = np.arange(0, num_entries, dtype=np.int64)
        pending: List["ObjectRef"] = []  # type: ignore[name-defined]
        results: Dict[int, Ephemeris] = {}

        for idx_chunk in _iterate_chunks(idx, chunk_size):
            start = int(idx_chunk[0]) if len(idx_chunk) else 0
            pending.append(
                ephemeris_2body_worker_ray.remote(  # type: ignore[name-defined]
                    start,
                    idx_chunk,
                    propagated_ref,
                    observers_ref,
                    lt_tol,
                    max_iter,
                    tol,
                    stellar_aberration,
                    predict_magnitudes,
                    predict_phase_angle,
                )
            )

            if len(pending) >= max_processes * 1.5:
                finished, pending = ray.wait(pending, num_returns=1)  # type: ignore[name-defined]
                start_i, eph_i = ray.get(finished[0])  # type: ignore[name-defined]
                results[int(start_i)] = eph_i

        while pending:
            finished, pending = ray.wait(pending, num_returns=1)  # type: ignore[name-defined]
            start_i, eph_i = ray.get(finished[0])  # type: ignore[name-defined]
            results[int(start_i)] = eph_i

        chunks = [results[k] for k in sorted(results.keys())]
        return qv.concatenate(chunks) if chunks else Ephemeris.empty()

    num_entries = len(observers)
    assert (
        len(propagated_orbits) == num_entries
    ), "Orbits and observers must be paired and orbits must be propagated to observer times."

    # Transform both the orbits and observers to the barycenter if they are not already.
    propagated_orbits_barycentric = propagated_orbits.set_column(
        "coordinates",
        transform_coordinates(
            propagated_orbits.coordinates,
            CartesianCoordinates,
            frame_out="ecliptic",
            origin_out=OriginCodes.SOLAR_SYSTEM_BARYCENTER,
        ),
    )
    observers_barycentric = observers.set_column(
        "coordinates",
        transform_coordinates(
            observers.coordinates,
            CartesianCoordinates,
            frame_out="ecliptic",
            origin_out=OriginCodes.SOLAR_SYSTEM_BARYCENTER,
        ),
    )

    observer_coordinates = observers_barycentric.coordinates.values
    observer_codes = observers_barycentric.code.to_numpy(zero_copy_only=False)
    mu = observers_barycentric.coordinates.origin.mu()
    times = propagated_orbits.coordinates.time.mjd().to_numpy(zero_copy_only=False)

    # Define chunk size
    chunk_size = 200

    # Process in chunks
    ephemeris_spherical = np.empty((num_entries, 6), dtype=np.float64)
    light_time = np.empty((num_entries,), dtype=np.float64)
    aberrated_orbits = np.empty((num_entries, 6), dtype=np.float64)
    start = 0
    for orbits_chunk, times_chunk, observer_coords_chunk, mu_chunk in zip(
        process_in_chunks(propagated_orbits_barycentric.coordinates.values, chunk_size),
        process_in_chunks(times, chunk_size),
        process_in_chunks(observer_coordinates, chunk_size),
        process_in_chunks(mu, chunk_size),
    ):
        valid = min(chunk_size, num_entries - start)
        ephemeris_chunk, light_time_chunk, aberrated_chunk = (
            _generate_ephemeris_2body_vmap(
                orbits_chunk,
                times_chunk,
                observer_coords_chunk,
                mu_chunk,
                lt_tol,
                max_iter,
                tol,
                stellar_aberration,
            )
        )
        ephemeris_spherical[start : start + valid] = np.asarray(ephemeris_chunk)[:valid]
        light_time[start : start + valid] = np.asarray(light_time_chunk)[:valid]
        aberrated_orbits[start : start + valid] = np.asarray(aberrated_chunk)[:valid]
        start += valid

    if start != num_entries:
        raise RuntimeError(
            f"Internal error: expected {num_entries} ephemeris rows, got {start}"
        )

    # Compute emission times by subtracting light-time (in days) from the observation times.
    emission_times = propagated_orbits_barycentric.coordinates.time.add_fractional_days(
        pa.array(-light_time)
    )
    aberrated_coordinates = CartesianCoordinates.from_kwargs(
        x=aberrated_orbits[:, 0],
        y=aberrated_orbits[:, 1],
        z=aberrated_orbits[:, 2],
        vx=aberrated_orbits[:, 3],
        vy=aberrated_orbits[:, 4],
        vz=aberrated_orbits[:, 5],
        time=emission_times,
        origin=Origin.from_kwargs(
            code=np.full(num_entries, OriginCodes.SOLAR_SYSTEM_BARYCENTER.name)
        ),
        frame="ecliptic",
    )

    if not propagated_orbits.coordinates.covariance.is_all_nan():

        cartesian_covariances = propagated_orbits.coordinates.covariance.to_matrix()
        covariances_spherical = transform_covariances_jacobian(
            propagated_orbits.coordinates.values,
            cartesian_covariances,
            _generate_ephemeris_2body,
            in_axes=(0, 0, 0, 0, None, None, None, None),
            out_axes=(0, 0, 0),
            observation_times=times,
            observer_coordinates=observer_coordinates,
            mu=mu,
            lt_tol=lt_tol,
            max_iter=max_iter,
            tol=tol,
            stellar_aberration=stellar_aberration,
        )
        covariances_spherical = CoordinateCovariances.from_matrix(
            np.array(covariances_spherical)
        )

    else:
        covariances_spherical = None

    spherical_coordinates = SphericalCoordinates.from_kwargs(
        time=propagated_orbits.coordinates.time,
        rho=ephemeris_spherical[:, 0],
        lon=ephemeris_spherical[:, 1],
        lat=ephemeris_spherical[:, 2],
        vrho=ephemeris_spherical[:, 3],
        vlon=ephemeris_spherical[:, 4],
        vlat=ephemeris_spherical[:, 5],
        covariance=covariances_spherical,
        origin=Origin.from_kwargs(code=observer_codes),
        frame="ecliptic",
    )

    # Rotate the spherical coordinates from the ecliptic frame
    # to the equatorial frame
    spherical_coordinates = transform_coordinates(
        spherical_coordinates, SphericalCoordinates, frame_out="equatorial"
    )

    ephemeris = Ephemeris.from_kwargs(
        orbit_id=propagated_orbits_barycentric.orbit_id,
        object_id=propagated_orbits_barycentric.object_id,
        coordinates=spherical_coordinates,
        light_time=light_time,
        aberrated_coordinates=aberrated_coordinates,
    )

    want_alpha = bool(predict_phase_angle)
    want_mags = bool(predict_magnitudes)

    if not want_alpha and not want_mags:
        return ephemeris

    # Determine whether we can compute magnitudes (needs H and G).
    has_params = None
    H_v = None
    G = None
    if want_mags:
        H_v = propagated_orbits.physical_parameters.H_v.to_numpy(zero_copy_only=False)
        G = propagated_orbits.physical_parameters.G.to_numpy(zero_copy_only=False)
        has_params = np.isfinite(H_v) & np.isfinite(G)
        if not np.any(has_params):
            want_mags = False

    if not want_alpha and not want_mags:
        return ephemeris

    # Transform object and observer coordinates to heliocentric for photometry.
    aberrated_heliocentric = transform_coordinates(
        aberrated_coordinates,
        CartesianCoordinates,
        frame_out="ecliptic",
        origin_out=OriginCodes.SUN,
    )
    observers_heliocentric = observers.set_column(
        "coordinates",
        transform_coordinates(
            observers.coordinates,
            CartesianCoordinates,
            frame_out="ecliptic",
            origin_out=OriginCodes.SUN,
        ),
    )

    alpha_deg = None
    mags = None
    if want_mags and want_alpha:
        assert H_v is not None and G is not None and has_params is not None
        mags, alpha_deg = calculate_apparent_magnitude_v_and_phase_angle(
            H_v=H_v,
            object_coords=aberrated_heliocentric,
            observer=observers_heliocentric,
            G=G,
        )
    elif want_alpha:
        alpha_deg = calculate_phase_angle(
            aberrated_heliocentric, observers_heliocentric
        )
    elif want_mags:
        assert H_v is not None and G is not None and has_params is not None
        mags = calculate_apparent_magnitude_v(
            H_v=H_v,
            object_coords=aberrated_heliocentric,
            observer=observers_heliocentric,
            G=G,
        )

    if alpha_deg is not None:
        alpha_deg = np.asarray(alpha_deg, dtype=np.float64)
        ephemeris = ephemeris.set_column(
            "alpha",
            pa.array(alpha_deg, mask=~np.isfinite(alpha_deg), type=pa.float64()),
        )

    if mags is not None:
        assert has_params is not None
        mags = np.asarray(mags, dtype=np.float64)
        valid = has_params & np.isfinite(mags)
        ephemeris = ephemeris.set_column(
            "predicted_magnitude_v", pa.array(mags, mask=~valid, type=pa.float64())
        )
    return ephemeris



import base64
from importlib import resources
from pathlib import Path

# Get the package resources for the logos directory
_logos_path = resources.files("adam_core.utils.plots.logos")

# Define paths to logo files
AsteroidInstituteLogoDark = Path(
    _logos_path / "AsteroidInstituteProgramForDark_Large.png"
)
AsteroidInstituteLogoLight = Path(
    _logos_path / "AsteroidInstituteProgramTransparent_Large.png"
)


def get_logo_base64(logo_path: Path) -> str:
    """Convert logo image to base64 string.

    Parameters
    ----------
    logo_path : Path
        Path to the logo image file

    Returns
    -------
    str
        Base64 encoded image data with data URL prefix
    """
    with open(logo_path, "rb") as f:
        encoded = base64.b64encode(f.read()).decode("utf-8")
    return f"data:image/png;base64,{encoded}"


__all__ = ["AsteroidInstituteLogoDark", "AsteroidInstituteLogoLight", "get_logo_base64"]



import numpy as np


def pad_to_fixed_size(array, target_shape, pad_value=0):
    """
    Pad an array to a fixed shape with a specified pad value.

    Parameters
    ----------
    array : array-like
        Array to pad
    target_shape : tuple
        Desired output shape
    pad_value : int or float, optional
        Value to use for padding, by default 0

    Returns
    -------
    padded_array : array-like
        Padded array with desired shape
    """
    pad_width = [(0, max(0, t - s)) for s, t in zip(array.shape, target_shape)]
    return np.pad(array, pad_width, constant_values=pad_value)


def process_in_chunks(array, chunk_size):
    """
    Yield chunks of the array with a fixed size, padding the last chunk if necessary.

    Parameters
    ----------
    array : array-like
        Array to process in chunks
    chunk_size : int
        Size of each chunk

    Yields
    ------
    chunk : array-like
        Array chunk of fixed size (padded if necessary)
    """
    n = array.shape[0]
    for i in range(0, n, chunk_size):
        chunk = array[i : i + chunk_size]
        if chunk.shape[0] < chunk_size:
            chunk = pad_to_fixed_size(chunk, (chunk_size,) + chunk.shape[1:])
        yield chunk



from __future__ import annotations

import pyarrow as pa
import quivr as qv


class BandpassCurves(qv.Table):
    """
    Canonical bandpass response curves.

    Each row represents a single filter response curve: wavelength (nm) and a
    normalized dimensionless throughput (0..1).
    """

    filter_id = qv.LargeStringColumn()
    instrument = qv.LargeStringColumn()
    band = qv.LargeStringColumn()
    wavelength_nm = qv.LargeListColumn(pa.float64())
    throughput = qv.LargeListColumn(pa.float64())
    source = qv.LargeStringColumn()


class ObservatoryBandMap(qv.Table):
    """
    Map (MPC observatory_code, reported band) -> canonical filter_id.

    `key` is a convenience column: f"{observatory_code}|{reported_band}".
    """

    observatory_code = qv.LargeStringColumn()
    reported_band = qv.LargeStringColumn()
    filter_id = qv.LargeStringColumn()
    key = qv.LargeStringColumn()


class AsteroidTemplates(qv.Table):
    """
    Asteroid reflectance templates, including fixed population mixes.

    `reflectance` is dimensionless and should be normalized at 550 nm.
    """

    template_id = qv.LargeStringColumn()
    wavelength_nm = qv.LargeListColumn(pa.float64())
    reflectance = qv.LargeListColumn(pa.float64())
    weight_C = qv.Float64Column()
    weight_S = qv.Float64Column()
    citation = qv.LargeStringColumn()


class TemplateBandpassIntegrals(qv.Table):
    """
    Precomputed template×filter integrals.

    The stored scalar is intended for photon-counting synthetic photometry:
        I = ∫ F_sun(λ) * R_ast(λ) * T(λ) * λ dλ
    where F_sun is the adopted solar spectrum, R_ast is reflectance, and T is throughput.
    """

    template_id = qv.LargeStringColumn()
    filter_id = qv.LargeStringColumn()
    integral_photon = qv.Float64Column()



"""Tests for the scout module."""

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

from ..scout import ScoutOrbit, scout_orbits_to_variant_orbits


def test_scout_orbits_to_variant_orbits():
    """Test that scout orbits are correctly converted to variant orbits."""
    # Create a mock scout orbits table
    scout_data = {
        "idx": [0, 1],
        "epoch": ["60000.0", "60000.0"],
        "ec": ["0.5", "0.51"],
        "qr": ["1.0", "1.01"],
        "tp": ["59000.0", "59000.0"],
        "om": ["10.0", "10.1"],
        "w": ["50.0", "50.1"],
        "inc": ["10.0", "10.1"],
        "H": ["20.0", "20.0"],
        "dca": ["0.1", "0.1"],
        "tca": ["0.1", "0.1"],
        "moid": ["0.1", "0.1"],
        "vinf": ["0.1", "0.1"],
        "geoEcc": ["0.1", "0.1"],
        "impFlag": [0, 0],
    }
    scout_orbits = ScoutOrbit.from_kwargs(**scout_data)

    # Convert to variant orbits
    variant_orbits = scout_orbits_to_variant_orbits("2024AA", scout_orbits)

    # Check that the output has the expected structure
    assert len(variant_orbits) == len(scout_orbits)
    assert variant_orbits.coordinates.frame == "ecliptic"
    assert pc.all(pc.equal(variant_orbits.coordinates.origin.code, "SUN")).as_py()

    # Check that the object IDs are correct
    assert variant_orbits.object_id.to_pylist() == ["2024AA", "2024AA"]

    # Check that the orbit IDs are correct
    assert variant_orbits.orbit_id.to_pylist() == ["0", "1"]

    # Check that the variant IDs are unique
    assert len(pc.unique(variant_orbits.variant_id)) == len(scout_orbits)

    # Check that the time is correct
    np.testing.assert_array_equal(
        variant_orbits.coordinates.time.jd(), pc.cast(scout_orbits.epoch, pa.float64())
    )



import logging
import multiprocessing as mp
import warnings
from typing import List, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
import quivr as qv
import ray
from astropy.time import Time

from adam_core.coordinates import CartesianCoordinates, transform_coordinates
from adam_core.coordinates.origin import Origin, OriginCodes
from adam_core.coordinates.spherical import SphericalCoordinates
from adam_core.coordinates.units import au_per_day_to_km_per_s
from adam_core.dynamics.lambert import calculate_c3, solve_lambert
from adam_core.orbits import Orbits
from adam_core.propagator import Propagator
from adam_core.ray_cluster import initialize_use_ray
from adam_core.time import Timestamp
from adam_core.utils import get_perturber_state
from adam_core.utils.iter import _iterate_chunk_indices
from adam_core.utils.plots.logos import AsteroidInstituteLogoLight, get_logo_base64

logger = logging.getLogger(__name__)


def generate_saturated_colorscale(
    base_color: str, n_levels: int = 8, max_alpha: float = 0.8, min_alpha: float = 0.1
) -> List[List]:
    """
    Generate a colorscale from light to dark based on a base color with full saturation
    and variable transparency that increases with color intensity.

    Parameters
    ----------
    base_color : str
        Base color name (e.g., 'red', 'blue') or hex code (e.g., '#FF0000')
    n_levels : int, optional
        Number of levels in the colorscale (default: 8)
    max_alpha : float, optional
        Maximum alpha (opacity) for darkest colors (default: 0.8)
    min_alpha : float, optional
        Minimum alpha (opacity) for lightest colors (default: 0.1)

    Returns
    -------
    List[List]
        Plotly colorscale format with RGBA: [[position, color], ...]
    """
    # Color mapping for common base colors to RGB with full saturation
    color_map = {
        "red": (255, 0, 0),  # Pure red, full saturation
        "blue": (0, 0, 255),  # Pure blue, full saturation
        "green": (0, 255, 0),  # Pure green, full saturation
        "orange": (255, 165, 0),  # Pure orange, full saturation
        "purple": (128, 0, 128),  # Pure purple, full saturation
        "yellow": (255, 255, 0),  # Pure yellow, full saturation
        "cyan": (0, 255, 255),  # Pure cyan, full saturation
        "magenta": (255, 0, 255),  # Pure magenta, full saturation
    }

    # Parse base color
    if base_color.startswith("#"):
        # Hex color
        hex_color = base_color.lstrip("#")
        base_rgb = tuple(int(hex_color[i : i + 2], 16) for i in (0, 2, 4))
    elif base_color.lower() in color_map:
        base_rgb = color_map[base_color.lower()]
    else:
        # Default to pure red if unknown
        base_rgb = (255, 0, 0)

    colorscale = []
    for i in range(n_levels):
        position = i / (n_levels - 1)

        # Create lightness variation while maintaining full saturation
        # Lighter colors: mix with white but preserve hue
        # Darker colors: reduce brightness but keep saturation
        if position == 0:
            # Lightest: mix with white for pastel effect
            lightness_factor = 0.9  # Very light
            r = int(base_rgb[0] * lightness_factor + 255 * (1 - lightness_factor))
            g = int(base_rgb[1] * lightness_factor + 255 * (1 - lightness_factor))
            b = int(base_rgb[2] * lightness_factor + 255 * (1 - lightness_factor))
        else:
            # Use power curve for smooth transition
            intensity = np.power(position, 0.8)

            # Maintain saturation by scaling from full saturation down
            r = int(base_rgb[0] * (0.3 + 0.7 * intensity))
            g = int(base_rgb[1] * (0.3 + 0.7 * intensity))
            b = int(base_rgb[2] * (0.3 + 0.7 * intensity))

        # Ensure values are within valid RGB range
        r = max(0, min(255, r))
        g = max(0, min(255, g))
        b = max(0, min(255, b))

        # Calculate alpha based on position (lighter = more transparent)
        alpha = min_alpha + (max_alpha - min_alpha) * position

        colorscale.append([position, f"rgba({r}, {g}, {b}, {alpha:.2f})"])

    return colorscale


def generate_perceptual_colorscale(
    base_color: str,
    n_levels: int = 8,
    min_lightness: float = 0.3,
    max_lightness: float = 0.9,
    max_alpha: float = 0.8,
    min_alpha: float = 0.1,
) -> List[List]:
    """
    Generate a perceptually uniform colorscale with full saturation and variable transparency
    that works better for overlaying contours.

    Parameters
    ----------
    base_color : str
        Base color name (e.g., 'red', 'blue') or hex code (e.g., '#FF0000')
    n_levels : int, optional
        Number of levels in the colorscale (default: 8)
    min_lightness : float, optional
        Minimum lightness value (0-1, default: 0.3 for good contrast)
    max_lightness : float, optional
        Maximum lightness value (0-1, default: 0.9 for visibility with transparency)
    max_alpha : float, optional
        Maximum alpha (opacity) for darkest colors (default: 0.8)
    min_alpha : float, optional
        Minimum alpha (opacity) for lightest colors (default: 0.1)

    Returns
    -------
    List[List]
        Plotly colorscale format with RGBA: [[position, color], ...]
    """
    # Full saturation color mapping for maximum color purity
    color_map = {
        "red": (255, 0, 0),  # Pure red
        "blue": (0, 0, 255),  # Pure blue
        "green": (0, 255, 0),  # Pure green
        "orange": (255, 165, 0),  # Pure orange
        "purple": (128, 0, 128),  # Pure purple
        "yellow": (255, 255, 0),  # Pure yellow
        "cyan": (0, 255, 255),  # Pure cyan
        "magenta": (255, 0, 255),  # Pure magenta
    }

    # Parse base color
    if base_color.startswith("#"):
        hex_color = base_color.lstrip("#")
        base_rgb = tuple(int(hex_color[i : i + 2], 16) for i in (0, 2, 4))
    elif base_color.lower() in color_map:
        base_rgb = color_map[base_color.lower()]
    else:
        base_rgb = (255, 0, 0)  # Default pure red

    # Convert base color to normalized RGB for calculations
    base_r, base_g, base_b = [x / 255.0 for x in base_rgb]

    colorscale = []
    for i in range(n_levels):
        position = i / (n_levels - 1)

        # Create perceptually uniform lightness steps
        lightness = max_lightness - (position * (max_lightness - min_lightness))

        # Use full saturation throughout, varying only lightness

        if position == 0:
            # Lightest: mix with white while maintaining hue
            white_mix = 1 - lightness
            r = base_r * lightness + white_mix
            g = base_g * lightness + white_mix
            b = base_b * lightness + white_mix
        else:
            # Scale the base color by lightness while maintaining saturation
            # Use HSV-like scaling to preserve hue and saturation
            max_component = max(base_r, base_g, base_b)

            if max_component > 0:
                # Scale all components proportionally to achieve desired lightness
                scale_factor = lightness / max_component
                r = base_r * scale_factor
                g = base_g * scale_factor
                b = base_b * scale_factor
            else:
                r = g = b = lightness

        # Convert back to 0-255 range and ensure validity
        r_int = max(0, min(255, int(r * 255)))
        g_int = max(0, min(255, int(g * 255)))
        b_int = max(0, min(255, int(b * 255)))

        # Calculate alpha based on position (lighter = more transparent)
        alpha = min_alpha + (max_alpha - min_alpha) * position

        colorscale.append([position, f"rgba({r_int}, {g_int}, {b_int}, {alpha:.2f})"])

    return colorscale


# class LambertOutput(qv.Table):
class LambertSolutions(qv.Table):
    # departure_state = CartesianCoordinates.as_column()
    # arrival_state = CartesianCoordinates.as_column()
    departure_body_id = qv.LargeStringColumn()
    departure_time = Timestamp.as_column()
    departure_body_x = qv.Float64Column()
    departure_body_y = qv.Float64Column()
    departure_body_z = qv.Float64Column()
    departure_body_vx = qv.Float64Column()
    departure_body_vy = qv.Float64Column()
    departure_body_vz = qv.Float64Column()
    arrival_body_id = qv.LargeStringColumn()
    arrival_time = Timestamp.as_column()
    arrival_body_x = qv.Float64Column()
    arrival_body_y = qv.Float64Column()
    arrival_body_z = qv.Float64Column()
    arrival_body_vx = qv.Float64Column()
    arrival_body_vy = qv.Float64Column()
    arrival_body_vz = qv.Float64Column()
    solution_departure_vx = qv.Float64Column()
    solution_departure_vy = qv.Float64Column()
    solution_departure_vz = qv.Float64Column()
    solution_arrival_vx = qv.Float64Column()
    solution_arrival_vy = qv.Float64Column()
    solution_arrival_vz = qv.Float64Column()
    frame = qv.StringAttribute(default="unspecified")
    origin = Origin.as_column()

    def departure_body_orbit(self) -> Orbits:
        """
        Return the departure body orbit.
        """
        return Orbits.from_kwargs(
            orbit_id=self.departure_body_id,
            coordinates=CartesianCoordinates.from_kwargs(
                time=self.departure_time,
                x=self.departure_body_x,
                y=self.departure_body_y,
                z=self.departure_body_z,
                vx=self.departure_body_vx,
                vy=self.departure_body_vy,
                vz=self.departure_body_vz,
                origin=self.origin,
                frame=self.frame,
            ),
        )

    def arrival_body_orbit(self) -> Orbits:
        """
        Return the arrival body orbit.
        """
        return Orbits.from_kwargs(
            orbit_id=self.arrival_body_id,
            coordinates=CartesianCoordinates.from_kwargs(
                time=self.arrival_time,
                x=self.arrival_body_x,
                y=self.arrival_body_y,
                z=self.arrival_body_z,
                vx=self.arrival_body_vx,
                vy=self.arrival_body_vy,
                vz=self.arrival_body_vz,
                origin=self.origin,
                frame=self.frame,
            ),
        )

    def solution_departure_orbit(self) -> Orbits:
        """
        Return the solution departure orbit.
        """
        solution_departure_orbit_id = [
            f"solution_departure_orbit_{i}"
            for i in range(len(self.solution_departure_vx))
        ]
        return Orbits.from_kwargs(
            orbit_id=solution_departure_orbit_id,
            coordinates=CartesianCoordinates.from_kwargs(
                time=self.departure_time,
                x=self.departure_body_x,
                y=self.departure_body_y,
                z=self.departure_body_z,
                vx=self.solution_departure_vx,
                vy=self.solution_departure_vy,
                vz=self.solution_departure_vz,
                origin=self.origin,
                frame=self.frame,
            ),
        )

    def solution_arrival_orbit(self) -> Orbits:
        """
        Return the solution arrival orbit.
        """
        solution_arrival_orbit_id = [
            f"solution_arrival_orbit_{i}" for i in range(len(self.solution_arrival_vx))
        ]
        return Orbits.from_kwargs(
            orbit_id=solution_arrival_orbit_id,
            coordinates=CartesianCoordinates.from_kwargs(
                time=self.arrival_time,
                x=self.arrival_body_x,
                y=self.arrival_body_y,
                z=self.arrival_body_z,
                vx=self.solution_arrival_vx,
                vy=self.solution_arrival_vy,
                vz=self.solution_arrival_vz,
                origin=self.origin,
                frame=self.frame,
            ),
        )

    def c3_departure(self) -> npt.NDArray[np.float64]:
        """
        Return the C3 in au^2/d^2.
        """
        return calculate_c3(
            np.array(
                self.table.select(
                    [
                        "solution_departure_vx",
                        "solution_departure_vy",
                        "solution_departure_vz",
                    ]
                )
            ),
            np.array(
                self.table.select(
                    ["departure_body_vx", "departure_body_vy", "departure_body_vz"]
                )
            ),
        )

    def c3_arrival(self) -> npt.NDArray[np.float64]:
        """
        Return the C3 in au^2/d^2.
        """
        return calculate_c3(
            np.array(
                self.table.select(
                    [
                        "solution_arrival_vx",
                        "solution_arrival_vy",
                        "solution_arrival_vz",
                    ]
                )
            ),
            np.array(
                self.table.select(
                    ["arrival_body_vx", "arrival_body_vy", "arrival_body_vz"]
                )
            ),
        )

    def vinf_departure(self) -> npt.NDArray[np.float64]:
        """
        Return the v infinity in au/d.
        """
        return np.linalg.norm(
            np.array(
                self.table.select(
                    [
                        "solution_departure_vx",
                        "solution_departure_vy",
                        "solution_departure_vz",
                    ]
                )
            )
            - np.array(
                self.table.select(
                    ["departure_body_vx", "departure_body_vy", "departure_body_vz"]
                )
            ),
            axis=1,
        )

    def vinf_arrival(self) -> npt.NDArray[np.float64]:
        """
        Return the v infinity in au/d.
        """
        return np.linalg.norm(
            np.array(
                self.table.select(
                    [
                        "solution_arrival_vx",
                        "solution_arrival_vy",
                        "solution_arrival_vz",
                    ]
                )
            )
            - np.array(
                self.table.select(
                    ["arrival_body_vx", "arrival_body_vy", "arrival_body_vz"]
                )
            ),
            axis=1,
        )

    def time_of_flight(self) -> npt.NDArray[np.float64]:
        """
        Return the time of flight in days.
        """
        return self.arrival_time.mjd().to_numpy(
            zero_copy_only=False
        ) - self.departure_time.mjd().to_numpy(zero_copy_only=False)


def departure_spherical_coordinates(
    departure_origin: OriginCodes,
    times: Timestamp,
    frame: str,
    vx: npt.NDArray[np.float64],
    vy: npt.NDArray[np.float64],
    vz: npt.NDArray[np.float64],
) -> SphericalCoordinates:
    """
    Return the spherical coordinates of the departure vector.

    Parameters
    ----------
    departure_origin : OriginCodes
        The origin of the departure and also the frame of the departure vectors.
    times : Timestamp
        The times of the departure vectors.
    frame : str
        The frame of the departure vectors.
    vx : npt.NDArray[np.float64]
        The x-component of the departure vectors.
    vy : npt.NDArray[np.float64]
        The y-component of the departure vectors.
    vz : npt.NDArray[np.float64]
        The z-component of the departure vectors.

    Returns
    -------
    SphericalCoordinates
        The spherical coordinates of the departure unit vectors.
        Can be used to express ra / dec of the departure direction.
    """
    assert (
        len(vx) == len(vy) == len(vz) == len(times)
    ), "All arrays must have the same length"
    assert len(vx) > 0, "At least one departure vector is required"

    # Create unit direction vectors from the velocity vectors
    # Normalize the velocity vectors to get direction only
    velocity_magnitude = np.sqrt(vx**2 + vy**2 + vz**2)
    direction_x = vx / velocity_magnitude
    direction_y = vy / velocity_magnitude
    direction_z = vz / velocity_magnitude

    # Create CartesianCoordinates with the direction as position (on unit sphere)
    # and zero velocity since we only care about the direction
    direction_coords = CartesianCoordinates.from_kwargs(
        time=times,
        x=direction_x,  # Unit vector pointing in velocity direction
        y=direction_y,
        z=direction_z,
        vx=np.zeros_like(vx),  # No velocity needed for direction
        vy=np.zeros_like(vy),
        vz=np.zeros_like(vz),
        # From our departing origin.
        origin=Origin.from_OriginCodes(departure_origin, size=len(vx)),
        frame=frame,
    )

    # Transform direction to equatorial frame for proper RA/Dec coordinates
    # These are inertial celestial coordinates, suitable for any departure origin
    spherical = transform_coordinates(
        direction_coords,
        SphericalCoordinates,
        frame_out="equatorial",
        origin_out=departure_origin,
    )
    return spherical


def lambert_worker(
    departure_orbits: Orbits,
    arrival_orbits: Orbits,
    propagation_origin: OriginCodes,
    prograde: bool = True,
    max_iter: int = 35,
    tol: float = 1e-10,
) -> LambertSolutions:
    # Extract coordinates from orbits
    departure_coordinates = departure_orbits.coordinates
    arrival_coordinates = arrival_orbits.coordinates

    r1 = departure_coordinates.r
    r2 = arrival_coordinates.r
    tof = arrival_coordinates.time.mjd().to_numpy(
        zero_copy_only=False
    ) - departure_coordinates.time.mjd().to_numpy(zero_copy_only=False)

    origins = Origin.from_OriginCodes(propagation_origin, size=len(r1))
    mu = origins.mu()[0]
    v1, v2 = solve_lambert(r1, r2, tof, mu, prograde, max_iter, tol)

    # Use actual orbit IDs from the Orbits objects
    departure_body_ids = departure_orbits.orbit_id.to_pylist()
    arrival_body_ids = arrival_orbits.orbit_id.to_pylist()

    return LambertSolutions.from_kwargs(
        departure_body_id=departure_body_ids,
        departure_time=departure_coordinates.time,
        departure_body_x=departure_coordinates.x,
        departure_body_y=departure_coordinates.y,
        departure_body_z=departure_coordinates.z,
        departure_body_vx=departure_coordinates.vx,
        departure_body_vy=departure_coordinates.vy,
        departure_body_vz=departure_coordinates.vz,
        arrival_body_id=arrival_body_ids,
        arrival_time=arrival_coordinates.time,
        arrival_body_x=arrival_coordinates.x,
        arrival_body_y=arrival_coordinates.y,
        arrival_body_z=arrival_coordinates.z,
        arrival_body_vx=arrival_coordinates.vx,
        arrival_body_vy=arrival_coordinates.vy,
        arrival_body_vz=arrival_coordinates.vz,
        solution_departure_vx=v1[:, 0],
        solution_departure_vy=v1[:, 1],
        solution_departure_vz=v1[:, 2],
        solution_arrival_vx=v2[:, 0],
        solution_arrival_vy=v2[:, 1],
        solution_arrival_vz=v2[:, 2],
        frame=departure_coordinates.frame,
        origin=origins,
    )


lambert_worker_remote = ray.remote(lambert_worker)


def prepare_and_propagate_orbits(
    body: Union[Orbits, OriginCodes],
    start_time: Timestamp,
    end_time: Timestamp,
    propagation_origin: OriginCodes = OriginCodes.SUN,
    step_size: float = 1.0,
    propagator_class: Optional[type[Propagator]] = None,
    max_processes: Optional[int] = 1,
) -> Orbits:
    """
    Prepare and propagate orbits for a single body over a specified time range.

    Parameters
    ----------
    body : Union[Orbits, OriginCodes]
        The body to propagate (either an Orbits object or an OriginCode for a major body).
    start_time : Timestamp
        The start time for propagation.
    end_time : Timestamp
        The end time for propagation.
    propagation_origin : OriginCodes, optional
        The origin of the propagation (default: SUN).
    step_size : float, optional
        The step size in days (default: 1.0).
    propagator_class : Optional[type[Propagator]], optional
        The propagator class to use for orbit propagation.
    max_processes : Optional[int], optional
        The maximum number of processes to use.

    Returns
    -------
    Orbits
        The propagated orbits over the specified time range.
    """
    # if body is an Orbit, ensure its origin is the propagation_origin
    if isinstance(body, Orbits):
        body = body.set_column(
            "coordinates",
            transform_coordinates(
                body.coordinates,
                representation_out=CartesianCoordinates,
                frame_out="ecliptic",
                origin_out=propagation_origin,
            ),
        )

    times = Timestamp.from_mjd(
        np.arange(
            start_time.rescale("tdb").mjd()[0].as_py(),
            end_time.rescale("tdb").mjd()[0].as_py(),
            step_size,
        ),
        scale="tdb",
    )

    # get orbits for the body at specified times
    if isinstance(body, Orbits):
        propagator = propagator_class()
        orbits = propagator.propagate_orbits(body, times, max_processes=max_processes)
    else:
        # For major bodies, create an Orbits object with the body's origin code as the orbit_id
        coordinates = get_perturber_state(
            body, times, frame="ecliptic", origin=propagation_origin
        )
        # Create orbit IDs based on the body name and time index
        orbit_ids = np.repeat(body.name, len(coordinates))
        orbits = Orbits.from_kwargs(
            orbit_id=orbit_ids,
            coordinates=coordinates,
        )

    return orbits


def generate_porkchop_data(
    departure_orbits: Orbits,
    arrival_orbits: Orbits,
    propagation_origin: OriginCodes = OriginCodes.SUN,
    prograde: bool = True,
    max_iter: int = 35,
    tol: float = 1e-10,
    max_processes: Optional[int] = 1,
) -> LambertSolutions:
    """
    Generate data for a porkchop plot by solving Lambert's problem for a grid of
    departure and arrival times.

    Parameters
    ----------
    departure_orbits : Orbits
        The departure orbits.
    arrival_orbits : Orbits
        The arrival orbits.
    propagation_origin : OriginCodes
        The origin of the propagation.
    prograde : bool, optional
        If True, assume prograde motion. If False, assume retrograde motion.
    max_iter : int, optional
        The maximum number of iterations for Lambert's solver.
    tol : float, optional
        The numerical tolerance for Lambert's solver.
    max_processes : Optional[int], optional
        The maximum number of processes to use.
    max_processes : Optional[int], optional
        The maximum number of processes to use.


    Returns
    -------
    porkchop_data : LambertOutput
        The porkchop data.
    """

    assert (
        departure_orbits.coordinates.frame == arrival_orbits.coordinates.frame
    ), "Departure and arrival frames must be the same"
    assert len(departure_orbits.coordinates.origin.code.unique()) == 1
    assert len(arrival_orbits.coordinates.origin.code.unique()) == 1

    assert (
        departure_orbits.coordinates.origin.code[0]
        == arrival_orbits.coordinates.origin.code[0]
    ), "Departure and arrival origins must be the same"

    # First let's make sure departure and arrival orbits are time-ordered
    departure_orbits = departure_orbits.sort_by(
        ["coordinates.time.days", "coordinates.time.nanos"]
    )
    arrival_orbits = arrival_orbits.sort_by(
        ["coordinates.time.days", "coordinates.time.nanos"]
    )

    # Get the actual times for comparison
    dep_times_mjd = departure_orbits.coordinates.time.mjd().to_numpy(
        zero_copy_only=False
    )
    arr_times_mjd = arrival_orbits.coordinates.time.mjd().to_numpy(zero_copy_only=False)

    # Create meshgrids of indices and times
    dep_indices, arr_indices = np.meshgrid(
        np.arange(len(departure_orbits)), np.arange(len(arrival_orbits))
    )
    dep_time_grid, arr_time_grid = np.meshgrid(dep_times_mjd, arr_times_mjd)

    # Filter to ensure departure time is before arrival time
    # Use actual time comparison instead of index comparison
    valid_indices = arr_time_grid > dep_time_grid

    # Apply the mask to flatten only valid combinations
    dep_indices_flat = dep_indices[valid_indices].flatten()
    arr_indices_flat = arr_indices[valid_indices].flatten()

    stacked_departure_orbits = departure_orbits.take(dep_indices_flat)
    stacked_arrival_orbits = arrival_orbits.take(arr_indices_flat)

    # If no valid combinations exist, return empty results
    if len(stacked_departure_orbits) == 0:
        return LambertSolutions.empty()

    if max_processes is None:
        max_processes = mp.cpu_count()

    use_ray = initialize_use_ray(max_processes)

    lambert_results = LambertSolutions.empty()
    if use_ray:
        futures = []
        for start, end in _iterate_chunk_indices(
            stacked_departure_orbits, chunk_size=100
        ):
            futures.append(
                lambert_worker_remote.remote(
                    stacked_departure_orbits[start:end],
                    stacked_arrival_orbits[start:end],
                    propagation_origin,
                    prograde,
                    max_iter,
                    tol,
                )
            )

            if len(futures) > 1.5 * max_processes:
                finished, futures = ray.wait(futures, num_returns=1)
                result = ray.get(finished[0])
                lambert_results = qv.concatenate([lambert_results, result])

        while futures:
            finished, futures = ray.wait(futures, num_returns=1)
            result = ray.get(finished[0])
            lambert_results = qv.concatenate([lambert_results, result])

    else:
        lambert_results = lambert_worker(
            stacked_departure_orbits,
            stacked_arrival_orbits,
            propagation_origin,
            prograde,
            max_iter,
            tol,
        )

    return lambert_results


def plot_porkchop_plotly(
    porkchop_data: LambertSolutions,
    width: int = 900,
    height: int = 700,
    c3_departure_min: Optional[float] = None,
    c3_departure_max: Optional[float] = None,
    vinf_arrival_min: Optional[float] = None,
    vinf_arrival_max: Optional[float] = None,
    tof_min: Optional[float] = None,
    tof_max: Optional[float] = None,
    c3_base_colorscale: str = "Reds",
    vinf_base_colorscale: str = "Blues",
    tof_line_color: str = "black",
    xlim_mjd: Optional[Tuple[float, float]] = None,
    ylim_mjd: Optional[Tuple[float, float]] = None,
    title: str = "Porkchop Plot",
    show_optimal: bool = True,
    show_hover: bool = False,
    logo: bool = True,
):
    """
    Plot the porkchop plot from Lambert trajectory data using Plotly.

    Parameters
    ----------
    porkchop_data : LambertOutput
        The porkchop data.
    width : int, optional
        The width of the plot.
    height : int, optional
        The height of the plot.
    c3_departure_min : float, optional
        The minimum C3 departure value.
    c3_departure_max : float, optional
        The maximum C3 departure value.
    vinf_arrival_min : float, optional
        The minimum V∞ arrival value.
    vinf_arrival_max : float, optional
        The maximum V∞ arrival value.
    tof_min : float, optional
        The minimum time of flight value.
    tof_max : float, optional
        The maximum time of flight value.
    c3_base_colorscale : str, optional
        The base colorscale for C3.
    vinf_base_colorscale : str, optional
        The base colorscale for V∞.
    tof_line_color : str, optional
        The color of the time of flight line.
    xlim_mjd : tuple, optional
        The x-axis limits in MJD.
    ylim_mjd : tuple, optional
        The y-axis limits in MJD.
    title : str, optional
        The title of the plot.
    show_optimal : bool, optional
        Whether to show the optimal V∞ point.
    show_hover : bool, optional
        Whether to show the hover information.
    logo : bool, optional
        Whether to show the logo.

    Returns
    -------
    fig : plotly.graph_objects.Figure
        The Plotly figure object.
    """
    # --- Extract basic raw data ---
    c3_departure_au_d2 = porkchop_data.c3_departure()  # C3 departure in (AU/day)^2
    vinf_arrival_au_day = porkchop_data.vinf_arrival()  # V∞ arrival in AU/day
    time_of_flight_days = porkchop_data.time_of_flight()
    departure_times = porkchop_data.departure_time
    arrival_times = porkchop_data.arrival_time

    # Convert to metric units using unit conversion functions
    c3_departure_km2_s2 = c3_departure_au_d2 * (au_per_day_to_km_per_s(1.0) ** 2)
    vinf_arrival_km_s = au_per_day_to_km_per_s(vinf_arrival_au_day)
    # Define default C3 range if not provided
    if c3_departure_min is None:
        c3_departure_min = 0
    if c3_departure_max is None:
        c3_departure_max = np.max(c3_departure_km2_s2)

    assert c3_departure_max > c3_departure_min, "C3 max must be greater than C3 min"

    c3_step = (c3_departure_max - c3_departure_min) / 10  # 10 levels by default
    assert c3_step < (
        c3_departure_max - c3_departure_min
    ), "C3 step must be less than the C3 range"

    # Define default V∞ range if not provided
    if vinf_arrival_min is None:
        vinf_arrival_min = 0
    if vinf_arrival_max is None:
        vinf_arrival_max = np.max(vinf_arrival_km_s)
    vinf_step = (vinf_arrival_max - vinf_arrival_min) / 10  # 10 levels by default

    if tof_min is None:
        tof_min = 0
    if tof_max is None:
        tof_max = np.max(time_of_flight_days)

    tof_step = max(5, (tof_max - tof_min) / 10)  # 10 levels, minimum step of 5 days
    tof_step = round(tof_step / 5) * 5  # Round to multiple of 5

    # Validate all step sizes are positive
    assert c3_step > 0, f"c3_step must be positive, got {c3_step}"
    assert vinf_step > 0, f"vinf_step must be positive, got {vinf_step}"
    assert tof_step > 0, f"tof_step must be positive, got {tof_step}"

    # Extract raw MJD values for all points
    departure_times_mjd = departure_times.mjd().to_numpy(zero_copy_only=False)
    arrival_times_mjd = arrival_times.mjd().to_numpy(zero_copy_only=False)

    # --- Apply all filtering to the actual data in one place ---
    # We want to keep all solutions that are not NaN and are within the specified ranges of c3, vinf and tof
    data_mask = (
        ~np.isnan(c3_departure_km2_s2)
        & ~np.isnan(vinf_arrival_km_s)  # Also filter out V∞ NaN values
        & (c3_departure_km2_s2 <= c3_departure_max)
        & (c3_departure_km2_s2 >= c3_departure_min)
        & (vinf_arrival_km_s >= vinf_arrival_min)
        & (vinf_arrival_km_s <= vinf_arrival_max)
        & (time_of_flight_days >= tof_min)
        & (time_of_flight_days <= tof_max)
    )

    # Filter all our data arrays using the combined mask
    filtered_departure_mjd = departure_times_mjd[data_mask]
    filtered_arrival_mjd = arrival_times_mjd[data_mask]
    filtered_c3_km2_s2 = c3_departure_km2_s2[data_mask]
    filtered_vinf_km_s = vinf_arrival_km_s[data_mask]
    filtered_tof_days = time_of_flight_days[data_mask]

    # Recalculate min/max and step sizes based on filtered data
    if len(filtered_c3_km2_s2) > 0:
        c3_departure_min_filtered = np.min(filtered_c3_km2_s2)
        c3_departure_max_filtered = np.max(filtered_c3_km2_s2)
        c3_step_filtered = (c3_departure_max_filtered - c3_departure_min_filtered) / 10
        if c3_step_filtered <= 0:
            c3_step_filtered = 1.0  # Fallback for constant data
    else:
        c3_departure_min_filtered = c3_departure_min
        c3_departure_max_filtered = c3_departure_max
        c3_step_filtered = c3_step

    if len(filtered_vinf_km_s) > 0:
        vinf_arrival_min_filtered = np.min(filtered_vinf_km_s)
        vinf_arrival_max_filtered = np.max(filtered_vinf_km_s)
        vinf_step_filtered = (
            vinf_arrival_max_filtered - vinf_arrival_min_filtered
        ) / 10
        if vinf_step_filtered <= 0:
            vinf_step_filtered = 1.0  # Fallback for constant data
    else:
        vinf_arrival_min_filtered = vinf_arrival_min
        vinf_arrival_max_filtered = vinf_arrival_max
        vinf_step_filtered = vinf_step

    if len(filtered_tof_days) > 0:
        tof_min_filtered = np.min(filtered_tof_days)
        tof_max_filtered = np.max(filtered_tof_days)
        tof_step_filtered = max(5, (tof_max_filtered - tof_min_filtered) / 10)
        tof_step_filtered = round(tof_step_filtered / 5) * 5  # Round to multiple of 5
        if tof_step_filtered <= 0:
            tof_step_filtered = 5  # Fallback minimum step
    else:
        tof_min_filtered = tof_min
        tof_max_filtered = tof_max
        tof_step_filtered = tof_step

    # Get unique times from the filtered data - this guarantees all data points have corresponding unique times
    unique_departure_mjd, dep_indices = np.unique(
        filtered_departure_mjd, return_inverse=True
    )
    unique_arrival_mjd, arr_indices = np.unique(
        filtered_arrival_mjd, return_inverse=True
    )
    # Check if we have enough unique times to create a grid
    if len(unique_departure_mjd) < 2 or len(unique_arrival_mjd) < 2:
        warnings.warn(
            "Porkchop plotting: Not enough unique times for grid. Returning empty figure."
        )
        fig_empty = go.Figure()
        fig_empty.update_layout(
            title=title + " (Insufficient data for grid)",
            xaxis_title="Departure Date",
            yaxis_title="Arrival Date",
            width=width,
            height=height,
            autosize=False,
            xaxis=dict(type="date"),
            yaxis=dict(type="date"),
        )
        return fig_empty

    # Convert to datetime objects for plotting axes
    unique_departure_dates_dt = [
        Time(mjd, format="mjd").datetime for mjd in unique_departure_mjd
    ]
    unique_arrival_dates_dt = [
        Time(mjd, format="mjd").datetime for mjd in unique_arrival_mjd
    ]

    # --- Unit Conversions and Grid Setup ---
    # Create the grid including date combinations that do not have valid Lambert solutions
    grid_departure_mjd, grid_arrival_mjd = np.meshgrid(
        unique_departure_mjd, unique_arrival_mjd
    )

    # Initialize grid arrays with NaN and fill using the filtered data
    # Since we used return_inverse=True, dep_indices and arr_indices are guaranteed to be valid
    grid_c3_departure_km2_s2 = np.full(
        (len(unique_arrival_mjd), len(unique_departure_mjd)), np.nan, dtype=np.float64
    )
    grid_vinf_arrival_km_s = np.full(
        (len(unique_arrival_mjd), len(unique_departure_mjd)), np.nan, dtype=np.float64
    )

    # Fill the grid directly - no validity masking needed since we pre-filtered the data
    grid_c3_departure_km2_s2[arr_indices, dep_indices] = filtered_c3_km2_s2
    grid_vinf_arrival_km_s[arr_indices, dep_indices] = filtered_vinf_km_s
    grid_tof_days = grid_arrival_mjd - grid_departure_mjd

    # --- Use original grids with NaN values for native Plotly handling ---
    grid_c3_for_plot = grid_c3_departure_km2_s2
    grid_vinf_for_plot = grid_vinf_arrival_km_s

    # Set up the date limits for the plot
    # Convert the min/max MJD values to datetime objects for Plotly
    xlim_dt = [
        Time(np.min(grid_departure_mjd), format="mjd").datetime,
        Time(np.max(grid_departure_mjd), format="mjd").datetime,
    ]
    ylim_dt = [
        Time(np.min(grid_arrival_mjd), format="mjd").datetime,
        Time(np.max(grid_arrival_mjd), format="mjd").datetime,
    ]

    # If explicit limits were provided, use those instead
    if xlim_mjd:
        xlim_dt = [
            Time(xlim_mjd[0], format="mjd").datetime,
            Time(xlim_mjd[1], format="mjd").datetime,
        ]
    if ylim_mjd:
        ylim_dt = [
            Time(ylim_mjd[0], format="mjd").datetime,
            Time(ylim_mjd[1], format="mjd").datetime,
        ]

    # --- Generate custom colorscales with better saturation at minimum values ---
    # Map common Plotly colorscale names to base colors
    colorscale_to_color = {
        "Reds": "red",
        "Blues": "blue",
        "Greens": "green",
        "Oranges": "orange",
        "Purples": "purple",
    }

    # Generate C3 colorscale with full saturation and built-in transparency
    if c3_base_colorscale in colorscale_to_color:
        # Using saturated colorscale with transparency built into the colorscale
        c3_colorscale = generate_saturated_colorscale(
            colorscale_to_color[c3_base_colorscale],
            n_levels=8,
            max_alpha=0.7,  # Maximum opacity for darkest colors
            min_alpha=0.15,  # Minimum opacity for lightest colors
        )
    else:
        c3_colorscale = c3_base_colorscale

    # Generate V∞ colorscale with full saturation and built-in transparency
    if vinf_base_colorscale in colorscale_to_color:
        # Using saturated colorscale with transparency built into the colorscale
        vinf_colorscale = generate_saturated_colorscale(
            colorscale_to_color[vinf_base_colorscale],
            n_levels=8,
            max_alpha=0.7,  # Maximum opacity for darkest colors
            min_alpha=0.15,  # Minimum opacity for lightest colors
        )
    else:
        vinf_colorscale = vinf_base_colorscale

    # --- Create hover information grids if requested ---
    hover_info = "none"
    custom_data = None
    hover_template = None

    if show_hover:
        # Create date strings for hover display
        grid_departure_date_strings = np.array(
            [
                [
                    Time(mjd, format="mjd").strftime("%Y-%m-%d")
                    for mjd in unique_departure_mjd
                ]
                for _ in unique_arrival_mjd
            ]
        )
        grid_arrival_date_strings = np.array(
            [
                [
                    Time(mjd, format="mjd").strftime("%Y-%m-%d")
                    for _ in unique_departure_mjd
                ]
                for mjd in unique_arrival_mjd
            ]
        )

        # Stack all the data we want in hover info
        # Shape: (n_arrival, n_departure, 5) for [c3, vinf, tof, dep_date, arr_date]
        custom_data = np.stack(
            [
                grid_c3_departure_km2_s2,  # C3 in km²/s²
                grid_vinf_arrival_km_s,  # V∞ in km/s
                grid_tof_days,  # ToF in days
                grid_departure_date_strings,  # Departure date strings
                grid_arrival_date_strings,  # Arrival date strings
            ],
            axis=-1,
        )

        hover_info = "text"
        hover_template = (
            "Departure: %{customdata[3]}<br>"
            "Arrival: %{customdata[4]}<br>"
            "Time of Flight: %{customdata[2]:.1f} days<br>"
            "C3 Departure: %{customdata[0]:.2f} km²/s²<br>"
            "V∞ Arrival: %{customdata[1]:.2f} km/s<br>"
            "<extra></extra>"
        )

    # --- Create Dual Contour Traces ---
    plotly_traces = []

    # C3 Departure Contour Trace (warm colorscale with built-in transparency)
    plotly_traces.append(
        go.Contour(
            x=unique_departure_dates_dt,
            y=unique_arrival_dates_dt,
            z=grid_c3_for_plot,
            zauto=False,
            zmin=c3_departure_min_filtered,
            zmax=c3_departure_max_filtered,
            colorscale=c3_colorscale,
            opacity=1.0,  # Use full opacity since transparency is built into colorscale
            hoverinfo=hover_info,
            hovertemplate=hover_template,
            customdata=custom_data,
            contours=dict(
                coloring="fill",
                showlabels=True,
                labelfont=dict(size=10, color="darkred"),
                start=c3_departure_min_filtered,
                end=c3_departure_max_filtered,
                size=c3_step_filtered,
                labelformat=".1f",
            ),
            ncontours=10,  # Ensure exactly 10 contour levels
            line=dict(width=1.0, smoothing=1.3),
            name="C3 Departure",
            showscale=False,  # Remove colorbar from main trace
            connectgaps=False,  # Don't connect across gaps to match V∞ behavior
            visible=True,
            showlegend=True,
        )
    )

    # V∞ Arrival Contour Trace (cool colorscale with built-in transparency)
    plotly_traces.append(
        go.Contour(
            x=unique_departure_dates_dt,
            y=unique_arrival_dates_dt,
            z=grid_vinf_for_plot,
            zauto=False,
            zmin=vinf_arrival_min_filtered,
            zmax=vinf_arrival_max_filtered,
            colorscale=vinf_colorscale,
            opacity=1.0,  # Use full opacity since transparency is built into colorscale
            hoverinfo=hover_info,
            hovertemplate=hover_template,
            customdata=custom_data,
            contours=dict(
                coloring="fill",
                showlabels=True,
                labelfont=dict(size=10, color="darkblue"),
                start=vinf_arrival_min_filtered,
                end=vinf_arrival_max_filtered,
                size=vinf_step_filtered,
                labelformat=".1f",
            ),
            ncontours=10,  # Ensure exactly 10 contour levels
            line=dict(width=1.0, smoothing=1.3),
            name="V∞ Arrival",
            showscale=False,  # Remove colorbar from main trace
            connectgaps=False,  # Faster rendering by not connecting across gaps
            visible="legendonly",
            showlegend=True,
        )
    )
    # --- ToF Contours ---
    plotly_traces.append(
        go.Contour(
            x=unique_departure_dates_dt,
            y=unique_arrival_dates_dt,
            z=grid_tof_days,  # Original ToF grid with NaNs
            colorscale=[[0, tof_line_color], [1, tof_line_color]],
            contours=dict(
                coloring="lines",
                showlabels=True,
                labelfont=dict(size=10, color=tof_line_color),
                start=tof_min_filtered,
                end=tof_max_filtered,
                size=tof_step_filtered,
            ),
            line=dict(color=tof_line_color, width=1, dash="longdash"),
            name="ToF (days)",
            showscale=False,
            hoverinfo="skip",  # Skip hover for ToF contours
            connectgaps=False,  # Don't connect across NaN gaps
            visible=True,
        )
    )

    # --- Optimal Points (separate for C3 and V∞) ---
    if show_optimal:
        # Optimal C3 Departure Point from filtered data
        if len(filtered_c3_km2_s2) > 0:
            min_c3_filtered_idx = np.nanargmin(filtered_c3_km2_s2)

            # Get the corresponding departure and arrival times from filtered data
            best_c3_dep_mjd = filtered_departure_mjd[min_c3_filtered_idx]
            best_c3_arr_mjd = filtered_arrival_mjd[min_c3_filtered_idx]

            # Convert to datetime objects for plotting
            best_c3_dep_dt = Time(best_c3_dep_mjd, format="mjd").datetime
            best_c3_arr_dt = Time(best_c3_arr_mjd, format="mjd").datetime

            # Check if the optimal C3 point falls within our current plot range
            c3_optimal_in_range = (
                xlim_dt[0] <= best_c3_dep_dt <= xlim_dt[1]
                and ylim_dt[0] <= best_c3_arr_dt <= ylim_dt[1]
            )

            if c3_optimal_in_range:
                plotly_traces.append(
                    go.Scatter(
                        x=[best_c3_dep_dt],
                        y=[best_c3_arr_dt],
                        mode="markers",
                        marker=dict(
                            symbol="circle",
                            color="darkred",
                            size=10,
                            line=dict(color="white", width=2),
                        ),
                        showlegend=True,
                        name="Optimal C3",
                        visible=True,
                        hoverinfo="skip",  # Skip hover for optimal points
                    )
                )

        # Optimal V∞ Arrival Point from filtered data
        if len(filtered_vinf_km_s) > 0:
            min_vinf_filtered_idx = np.nanargmin(filtered_vinf_km_s)

            # Get the corresponding departure and arrival times from filtered data
            best_vinf_dep_mjd = filtered_departure_mjd[min_vinf_filtered_idx]
            best_vinf_arr_mjd = filtered_arrival_mjd[min_vinf_filtered_idx]

            # Convert to datetime objects for plotting
            best_vinf_dep_dt = Time(best_vinf_dep_mjd, format="mjd").datetime
            best_vinf_arr_dt = Time(best_vinf_arr_mjd, format="mjd").datetime

            # Check if the optimal V∞ point falls within our current plot range
            vinf_optimal_in_range = (
                xlim_dt[0] <= best_vinf_dep_dt <= xlim_dt[1]
                and ylim_dt[0] <= best_vinf_arr_dt <= ylim_dt[1]
            )

            if vinf_optimal_in_range:
                plotly_traces.append(
                    go.Scatter(
                        x=[best_vinf_dep_dt],
                        y=[best_vinf_arr_dt],
                        mode="markers",
                        marker=dict(
                            symbol="circle",
                            color="darkblue",
                            size=10,
                            line=dict(color="white", width=2),
                        ),
                        showlegend=True,
                        name="Optimal V∞",
                        visible=True,
                        hoverinfo="skip",  # Skip hover for optimal points
                    )
                )

    # --- Figure Creation and Layout Update ---
    fig = go.Figure(data=plotly_traces)

    if logo:
        images = [
            dict(
                source=get_logo_base64(AsteroidInstituteLogoLight),
                xref="paper",
                yref="paper",
                x=0.98,
                y=0.02,
                sizex=0.12,
                sizey=0.12,
                xanchor="right",
                yanchor="bottom",
            )
        ]
    else:
        images = []

    fig.update_layout(
        title_text=title,
        xaxis_title="Departure Date",
        yaxis_title="Arrival Date",
        xaxis=dict(
            tickformat="%Y-%m-%d",
            tickangle=-45,
            range=xlim_dt,
            showgrid=True,
            gridcolor="lightgray",
            gridwidth=1,
        ),
        yaxis=dict(
            tickformat="%Y-%m-%d",
            range=ylim_dt,
            showgrid=True,
            gridcolor="lightgray",
            gridwidth=1,
        ),
        plot_bgcolor="white",
        width=width,
        height=height,
        autosize=False,
        hovermode="closest",
        images=images,
        legend=dict(
            orientation="h",
            yanchor="bottom",
            y=1.02,
            xanchor="left",
            x=0,
            itemsizing="constant",  # Keep legend items same size when hidden
            font=dict(size=12),  # Larger legend text
            bgcolor="rgba(255,255,255,0.8)",  # Semi-transparent background
            bordercolor="Black",
            borderwidth=1,
        ),
    )

    return fig



import logging
import multiprocessing as mp
from abc import ABC, abstractmethod
from typing import List, Literal, Optional, Type, Union

import numpy as np
import numpy.typing as npt
import pyarrow as pa
import pyarrow.compute as pc
import quivr as qv
import ray
from ray import ObjectRef

from ..constants import Constants as c
from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.origin import Origin, OriginCodes
from ..coordinates.spherical import SphericalCoordinates
from ..coordinates.transform import transform_coordinates
from ..dynamics.aberrations import add_light_time
from ..orbits.ephemeris import Ephemeris
from ..orbits.orbits import Orbits
from ..orbits.variants import VariantEphemeris, VariantOrbits
from ..photometry.magnitude import (
    calculate_apparent_magnitude_v,
    calculate_apparent_magnitude_v_and_phase_angle,
    calculate_phase_angle,
)
from ..ray_cluster import initialize_use_ray
from ..time import Timestamp
from ..utils.chunking import process_in_chunks
from ..utils.iter import _iterate_chunks
from .types import EphemerisType, ObserverType, OrbitType, TimestampType
from .utils import ensure_input_origin_and_frame, ensure_input_time_scale

logger = logging.getLogger(__name__)

C = c.C


def _alignment_indices_string_keys(
    *,
    obs_code: npt.NDArray[np.object_],
    obs_days: npt.NDArray[np.int64],
    obs_nanos: npt.NDArray[np.int64],
    eph_code: npt.NDArray[np.object_],
    eph_days: npt.NDArray[np.int64],
    eph_nanos: npt.NDArray[np.int64],
) -> npt.NDArray[np.int64]:
    keys_obs = np.array(
        [f"{c}|{d}|{n}" for c, d, n in zip(obs_code, obs_days, obs_nanos)],
        dtype=object,
    )
    keys_eph = np.array(
        [f"{c}|{d}|{n}" for c, d, n in zip(eph_code, eph_days, eph_nanos)],
        dtype=object,
    )
    if len(np.unique(keys_obs)) != len(keys_obs):
        raise ValueError(
            "Observer keys are not unique; cannot map ephemeris rows unambiguously."
        )

    sorter = np.argsort(keys_obs)
    pos = np.searchsorted(keys_obs[sorter], keys_eph)
    idx = sorter[pos]
    if np.any(keys_obs[idx] != keys_eph):
        raise ValueError("Failed to align ephemeris rows to observers (key mismatch).")
    return np.asarray(idx, dtype=np.int64)


def _alignment_indices_struct_index_in(
    *,
    obs_code: pa.Array,
    obs_days: pa.Array,
    obs_nanos: pa.Array,
    eph_code: pa.Array,
    eph_days: pa.Array,
    eph_nanos: pa.Array,
) -> npt.NDArray[np.int64]:
    # NOTE: Arrow compute `index_in` does not support struct kernels in our supported
    # versions, so we build keys via vectorized Arrow string concatenation (no Python loops).
    sep = pa.scalar("|", type=pa.large_string())
    obs_code_s = pc.cast(obs_code, pa.large_string())
    obs_days_s = pc.cast(obs_days, pa.large_string())
    obs_nanos_s = pc.cast(obs_nanos, pa.large_string())
    eph_code_s = pc.cast(eph_code, pa.large_string())
    eph_days_s = pc.cast(eph_days, pa.large_string())
    eph_nanos_s = pc.cast(eph_nanos, pa.large_string())

    obs_key = pc.binary_join_element_wise(
        pc.binary_join_element_wise(obs_code_s, obs_days_s, sep), obs_nanos_s, sep
    )
    eph_key = pc.binary_join_element_wise(
        pc.binary_join_element_wise(eph_code_s, eph_days_s, sep), eph_nanos_s, sep
    )

    # Uniqueness check: keys must uniquely identify an observer row.
    if int(pc.count_distinct(obs_key).as_py()) != len(obs_key):
        raise ValueError(
            "Observer keys are not unique; cannot map ephemeris rows unambiguously."
        )

    idx = pc.fill_null(pc.index_in(eph_key, value_set=obs_key), -1)
    idx_np = np.asarray(idx.to_numpy(zero_copy_only=False), dtype=np.int64)
    if np.any(idx_np < 0):
        raise ValueError("Failed to align ephemeris rows to observers (key mismatch).")
    return idx_np


def _hg_params_for_ephemeris_rows_arrow(
    *,
    orbit_id: pa.Array | pa.ChunkedArray,
    H_v: pa.Array | pa.ChunkedArray,
    G: pa.Array | pa.ChunkedArray,
    ephemeris_orbit_id: pa.Array | pa.ChunkedArray,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Vectorized mapping from orbit_id -> (H_v, G) aligned to ephemeris rows.

    This raises immediately if the input orbit table contains conflicting (H_v, G)
    values for the same orbit_id.
    """
    oid = pc.cast(orbit_id, pa.large_string())
    H = pc.cast(H_v, pa.float64())
    g = pc.cast(G, pa.float64())

    valid = pc.and_(pc.is_finite(H), pc.is_finite(g))
    if not pc.any(valid).as_py():
        n = len(ephemeris_orbit_id)
        return np.full(n, np.nan, dtype=np.float64), np.full(
            n, np.nan, dtype=np.float64
        )

    t = pa.table({"orbit_id": oid, "H_v": H, "G": g})
    t = t.filter(valid)

    gb = t.group_by("orbit_id").aggregate(
        [("H_v", "min"), ("H_v", "max"), ("G", "min"), ("G", "max")]
    )
    # Any mismatch indicates conflicting params for an orbit_id.
    conflict = pc.or_(
        pc.not_equal(gb["H_v_min"], gb["H_v_max"]),
        pc.not_equal(gb["G_min"], gb["G_max"]),
    )
    if pc.any(conflict).as_py():
        bad = pc.filter(gb["orbit_id"], conflict)
        bad_id = bad[0].as_py() if len(bad) > 0 else "<unknown>"
        raise ValueError(f"Conflicting physical parameters for orbit_id={bad_id}")

    # Map ephemeris orbit_id -> grouped index
    eph_oid = pc.cast(ephemeris_orbit_id, pa.large_string())
    idx = pc.index_in(eph_oid, value_set=gb["orbit_id"])
    # pc.take propagates nulls for missing indices; those convert to NaN below.
    H_out = pc.take(gb["H_v_min"], idx)
    G_out = pc.take(gb["G_min"], idx)
    return (
        np.asarray(H_out.to_numpy(zero_copy_only=False), dtype=np.float64),
        np.asarray(G_out.to_numpy(zero_copy_only=False), dtype=np.float64),
    )


def propagation_worker(
    orbits: Union[Orbits, VariantOrbits],
    times: Timestamp,
    propagator: Type["Propagator"],
    **kwargs,
) -> Union[Orbits, VariantOrbits]:
    prop = propagator(**kwargs)
    propagated = prop._propagate_orbits(orbits, times)
    return propagated


def attach_magnitude_or_phase(
    ephemeris: EphemerisType,
    orbits: OrbitType,
    observers: ObserverType,
    predict_magnitudes: bool,
    predict_phase_angle: bool,
) -> EphemerisType:
    if (not predict_magnitudes and not predict_phase_angle) or len(ephemeris) == 0:
        return ephemeris

    if pc.all(pc.is_null(ephemeris.aberrated_coordinates.x)).as_py():
        return ephemeris

    orbits_value = ray.get(orbits) if isinstance(orbits, ObjectRef) else orbits
    observers_value = (
        ray.get(observers) if isinstance(observers, ObjectRef) else observers
    )

    want_mags = bool(predict_magnitudes)
    want_alpha = bool(predict_phase_angle)

    H_v: npt.NDArray[np.float64] | None = None
    G: npt.NDArray[np.float64] | None = None
    has_params: npt.NDArray[np.bool_] | None = None
    if want_mags:
        H_v, G = _hg_params_for_ephemeris_rows_arrow(
            orbit_id=orbits_value.orbit_id,
            H_v=orbits_value.physical_parameters.H_v,
            G=orbits_value.physical_parameters.G,
            ephemeris_orbit_id=ephemeris.orbit_id,
        )
        has_params = np.isfinite(H_v) & np.isfinite(G)
        if not np.any(has_params):
            want_mags = False

    if not want_alpha and not want_mags:
        return ephemeris

    if ephemeris.coordinates.time.scale != observers_value.coordinates.time.scale:
        observers_value = observers_value.set_column(
            "coordinates.time",
            observers_value.coordinates.time.rescale(ephemeris.coordinates.time.scale),
        )

    idx_np = _alignment_indices_struct_index_in(
        obs_code=observers_value.code,
        obs_days=observers_value.coordinates.time.days,
        obs_nanos=observers_value.coordinates.time.nanos,
        eph_code=ephemeris.coordinates.origin.code,
        eph_days=ephemeris.coordinates.time.days,
        eph_nanos=ephemeris.coordinates.time.nanos,
    )
    observers_aligned = observers_value.take(pa.array(idx_np, type=pa.int64()))
    obj_helio = transform_coordinates(
        ephemeris.aberrated_coordinates,
        CartesianCoordinates,
        frame_out="ecliptic",
        origin_out=OriginCodes.SUN,
    )
    obs_helio = observers_aligned.set_column(
        "coordinates",
        transform_coordinates(
            observers_aligned.coordinates,
            CartesianCoordinates,
            frame_out="ecliptic",
            origin_out=OriginCodes.SUN,
        ),
    )

    alpha_deg = None
    mags = None
    if want_mags and want_alpha:
        assert H_v is not None and G is not None and has_params is not None
        mags, alpha_deg = calculate_apparent_magnitude_v_and_phase_angle(
            H_v=H_v,
            object_coords=obj_helio,
            observer=obs_helio,
            G=G,
        )
    elif want_alpha:
        alpha_deg = calculate_phase_angle(obj_helio, obs_helio)
    elif want_mags:
        assert H_v is not None and G is not None and has_params is not None
        mags = calculate_apparent_magnitude_v(
            H_v=H_v,
            object_coords=obj_helio,
            observer=obs_helio,
            G=G,
        )

    if alpha_deg is not None:
        alpha_deg = np.asarray(alpha_deg, dtype=np.float64)
        ephemeris = ephemeris.set_column(
            "alpha",
            pa.array(alpha_deg, mask=~np.isfinite(alpha_deg), type=pa.float64()),
        )

    if mags is not None:
        assert has_params is not None
        mags = np.asarray(mags, dtype=np.float64)
        valid = has_params & np.isfinite(mags)
        predicted = pa.array(mags, mask=~valid, type=pa.float64())
        ephemeris = ephemeris.set_column("predicted_magnitude_v", predicted)

    return ephemeris


def _uses_default_ephemeris_mixin(propagator: "Propagator") -> bool:
    """
    True iff `propagator._generate_ephemeris` is the EphemerisMixin implementation.

    Why: subclasses may override `_generate_ephemeris` without accepting extra keyword
    arguments. We only pass photometry flags into the default mixin implementation.
    """
    func = getattr(propagator._generate_ephemeris, "__func__", None)
    return func is EphemerisMixin._generate_ephemeris


@ray.remote
def propagation_worker_ray(
    idx: npt.NDArray[np.int64],
    orbits: OrbitType,
    times: OrbitType,
    propagator: "Propagator",
) -> OrbitType:
    orbits_chunk = orbits.take(idx)
    propagated = propagator._propagate_orbits(orbits_chunk, times)
    return propagated


@ray.remote
def ephemeris_worker_ray(
    idx: npt.NDArray[np.int64],
    orbits: OrbitType,
    observers: ObserverType,
    propagator: "Propagator",
    predict_magnitudes: bool,
    predict_phase_angle: bool,
) -> EphemerisType:
    orbits_chunk = orbits.take(idx)
    if _uses_default_ephemeris_mixin(propagator):
        ephemeris = propagator._generate_ephemeris(
            orbits_chunk,
            observers,
            predict_magnitudes=predict_magnitudes,
            predict_phase_angle=predict_phase_angle,
        )
    else:
        ephemeris = propagator._generate_ephemeris(orbits_chunk, observers)
        ephemeris = attach_magnitude_or_phase(
            ephemeris,
            orbits=orbits_chunk,
            observers=observers,
            predict_magnitudes=predict_magnitudes,
            predict_phase_angle=predict_phase_angle,
        )
    return ephemeris


class EphemerisMixin:
    """
    Mixin with signature for generating ephemerides.
    Subclasses should implement the _generate_ephemeris method.
    """

    def _generate_ephemeris(
        self,
        orbits: OrbitType,
        observers: ObserverType,
        *,
        predict_magnitudes: bool = True,
        predict_phase_angle: bool = False,
    ) -> EphemerisType:
        """
        A generic ephemeris implementation, which can be used or overridden by subclasses.
        """
        # Sort observers by time and code to ensure consistent ordering
        # As further propagation will order by time as well
        observers = observers.sort_by(
            ["coordinates.time.days", "coordinates.time.nanos", "code"]
        )

        observers_barycentric = observers.set_column(
            "coordinates",
            transform_coordinates(
                observers.coordinates,
                CartesianCoordinates,
                frame_out="ecliptic",
                origin_out=OriginCodes.SOLAR_SYSTEM_BARYCENTER,
            ),
        )
        observers_barycentric_tiled = qv.concatenate(
            [observers_barycentric] * len(orbits)
        )

        # Propagate orbits to sorted observer times
        # Returns orbits sorted by orbit_id and time
        propagated_orbits = self.propagate_orbits(
            orbits, observers.coordinates.time, max_processes=1
        )

        # Transform both the orbits and observers to the barycenter if they are not already.
        propagated_orbits_barycentric = propagated_orbits.set_column(
            "coordinates",
            transform_coordinates(
                propagated_orbits.coordinates,
                CartesianCoordinates,
                frame_out="ecliptic",
                origin_out=OriginCodes.SOLAR_SYSTEM_BARYCENTER,
            ),
        )

        # Process in padded chunks
        propagated_orbits_aberrated: np.ndarray = np.empty((0, 6))
        light_time: np.ndarray = np.empty((0,))

        propagated_orbits_barycentric_values = (
            propagated_orbits_barycentric.coordinates.values
        )
        propagated_orbits_barycentric_time = (
            propagated_orbits_barycentric.coordinates.time.mjd().to_numpy(
                zero_copy_only=False
            )
        )
        observers_barycentric_tiled_values = observers_barycentric_tiled.coordinates.r

        chunk_size = 200
        for (
            propagated_orbits_barycentric_chunk,
            propagated_orbits_barycentric_time_chunk,
            observers_barycentric_tiled_chunk,
        ) in zip(
            process_in_chunks(propagated_orbits_barycentric_values, chunk_size),
            process_in_chunks(propagated_orbits_barycentric_time, chunk_size),
            process_in_chunks(observers_barycentric_tiled_values, chunk_size),
        ):
            propagated_orbits_aberrated_chunk, light_time_chunk = add_light_time(
                propagated_orbits_barycentric_chunk,
                propagated_orbits_barycentric_time_chunk,
                observers_barycentric_tiled_chunk,
                lt_tol=1e-12,
                mu=c.MU,
                max_iter=100,
                tol=1e-15,
            )
            propagated_orbits_aberrated = np.concatenate(
                (
                    propagated_orbits_aberrated,
                    np.array(propagated_orbits_aberrated_chunk),
                )
            )
            light_time = np.concatenate((light_time, np.array(light_time_chunk)))

        # Remove padding
        propagated_orbits_aberrated = propagated_orbits_aberrated[
            : len(propagated_orbits_barycentric)
        ]
        light_time = light_time[: len(propagated_orbits_barycentric)]

        # Guard against pathological light-time values before constructing timestamps.
        if not np.all(np.isfinite(light_time)):
            raise ValueError(
                "Light travel time is NaN or too large and propagation will break."
            )

        # Compute emission times by subtracting light-time (in days) from the
        # propagated (observer) times.
        emission_times = (
            propagated_orbits_barycentric.coordinates.time.add_fractional_days(
                pa.array(-light_time)
            )
        )

        propagated_orbits_aberrated = Orbits.from_kwargs(
            orbit_id=propagated_orbits_barycentric.orbit_id,
            object_id=propagated_orbits_barycentric.object_id,
            coordinates=CartesianCoordinates.from_kwargs(
                x=propagated_orbits_aberrated[:, 0],
                y=propagated_orbits_aberrated[:, 1],
                z=propagated_orbits_aberrated[:, 2],
                vx=propagated_orbits_aberrated[:, 3],
                vy=propagated_orbits_aberrated[:, 4],
                vz=propagated_orbits_aberrated[:, 5],
                covariance=propagated_orbits_barycentric.coordinates.covariance,
                time=emission_times,
                origin=propagated_orbits_barycentric.coordinates.origin,
                frame=propagated_orbits_barycentric.coordinates.frame,
            ),
        )

        topocentric_state = (
            propagated_orbits_aberrated.coordinates.values
            - observers_barycentric_tiled.coordinates.values
        )
        topocentric_coordinates = CartesianCoordinates.from_kwargs(
            x=topocentric_state[:, 0],
            y=topocentric_state[:, 1],
            z=topocentric_state[:, 2],
            vx=topocentric_state[:, 3],
            vy=topocentric_state[:, 4],
            vz=topocentric_state[:, 5],
            covariance=None,
            # The ephemeris times are at the point of the observer,
            # not the aberrated orbit
            time=observers_barycentric_tiled.coordinates.time,
            origin=Origin.from_kwargs(
                code=observers_barycentric_tiled.code.to_numpy(zero_copy_only=False)
            ),
            frame="ecliptic",
        )

        spherical_coordinates = SphericalCoordinates.from_cartesian(
            topocentric_coordinates
        )

        spherical_coordinates = transform_coordinates(
            spherical_coordinates, SphericalCoordinates, frame_out="equatorial"
        )

        # Ephemeris are generally compared in UTC, so rescale the time
        spherical_coordinates = spherical_coordinates.set_column(
            "time",
            spherical_coordinates.time.rescale("utc"),
        )

        if isinstance(orbits, Orbits):
            ephemeris = Ephemeris.from_kwargs(
                orbit_id=propagated_orbits_barycentric.orbit_id,
                object_id=propagated_orbits_barycentric.object_id,
                coordinates=spherical_coordinates,
                light_time=light_time,
                aberrated_coordinates=propagated_orbits_aberrated.coordinates,
            )
        elif isinstance(orbits, VariantOrbits):

            # Propagated order is (orbit_id, variant_id, time) to match observer tiling
            # [observers]*len(orbits). Sort input the same way and repeat each variant's
            # weight for each time so row i gets the weight for the variant at that row.
            orbits_sorted = orbits.sort_by(
                [
                    "orbit_id",
                    "variant_id",
                    "coordinates.time.days",
                    "coordinates.time.nanos",
                ]
            )

            n_obs = len(observers)
            weights_np = orbits_sorted.weights.to_numpy(zero_copy_only=False)
            weights_cov_np = orbits_sorted.weights_cov.to_numpy(zero_copy_only=False)
            ephemeris = VariantEphemeris.from_kwargs(
                orbit_id=propagated_orbits_barycentric.orbit_id,
                object_id=propagated_orbits_barycentric.object_id,
                variant_id=propagated_orbits_barycentric.variant_id,
                coordinates=spherical_coordinates,
                light_time=light_time,
                weights=np.repeat(weights_np, n_obs),
                weights_cov=np.repeat(weights_cov_np, n_obs),
                aberrated_coordinates=propagated_orbits_aberrated.coordinates,
            )

        # Return in same order as propagate_orbits: (orbit_id, variant_id, time) or
        # (orbit_id, time). No sort here; callers use .sort_by() or .select()/.apply_mask().
        ephemeris = attach_magnitude_or_phase(
            ephemeris,
            orbits=orbits,
            observers=observers,
            predict_magnitudes=predict_magnitudes,
            predict_phase_angle=predict_phase_angle,
        )
        return ephemeris

    def generate_ephemeris(
        self,
        orbits: OrbitType,
        observers: ObserverType,
        covariance: bool = False,
        covariance_method: Literal[
            "auto", "sigma-point", "monte-carlo"
        ] = "monte-carlo",
        num_samples: int = 1000,
        chunk_size: int = 100,
        max_processes: Optional[int] = 1,
        seed: Optional[int] = None,
        predict_magnitudes: bool = True,
        predict_phase_angle: bool = False,
    ) -> Ephemeris:
        """
        Generate ephemerides for each orbit in orbits as observed by each observer
        in observers.

        Parameters
        ----------
        orbits : `~adam_core.orbits.orbits.Orbits` (N)
            Orbits for which to generate ephemerides.
        observers : `~adam_core.observers.observers.Observers` (M)
            Observers for which to generate the ephemerides of each
            orbit.
        covariance: bool, optional
            Propagate the covariance matrices of the orbits. This is done by sampling the
            orbits from their covariance matrices and propagating each sample and for each
            sample also generating ephemerides. The covariance
            of the ephemerides is then the covariance of the samples.
        covariance_method : {'sigma-point', 'monte-carlo', 'auto'}, optional
            The method to use for sampling the covariance matrix. If 'auto' is selected then the method
            will be automatically selected based on the covariance matrix. The default is 'monte-carlo'.
        num_samples : int, optional
            The number of samples to draw when sampling with monte-carlo.
        chunk_size : int, optional
            Number of orbits to send to each job.
        max_processes : int or None, optional
            Number of processes to launch. If None then the number of
            processes will be equal to the number of cores on the machine. If 1
            then no multiprocessing will be used. If "ray" is the parallel_backend and a ray instance
            is initialized already then this argument is ignored.

        Returns
        -------
        ephemeris : `~adam_core.orbits.ephemeris.Ephemeris` or `~adam_core.orbits.variants.VariantEphemeris`
            Predicted ephemerides. Row order matches :meth:`propagate_orbits`: (orbit_id, time)
            or (orbit_id, variant_id, time) for variant ephemeris. Use .sort_by() or
            .select()/.apply_mask() for other orderings or grouping.
        """
        # If sending in VariantOrbits, we make sure not to run covariance
        assert (covariance is False) or (
            isinstance(orbits, Orbits)
        ), "Covariance is not supported for VariantOrbits"

        # Check if we need to propagate orbit variants so we can propagate covariance
        # matrices
        ephemeris: Ephemeris = Ephemeris.empty()
        variant_ephemeris: VariantEphemeris = VariantEphemeris.empty()

        variants = VariantOrbits.empty()
        if covariance is True and not orbits.coordinates.covariance.is_all_nan():
            variants = VariantOrbits.create(
                orbits,
                method=covariance_method,
                num_samples=num_samples,
                seed=seed,
            )

        if max_processes is None:
            max_processes = mp.cpu_count()

        uses_default_ephemeris = _uses_default_ephemeris_mixin(self)
        needs_attach = max_processes <= 1 and not uses_default_ephemeris

        if max_processes > 1:
            initialize_use_ray(num_cpus=max_processes)

            # Add orbits and observers to object store if
            # they haven't already been added
            if not isinstance(observers, ObjectRef):
                observers_ref = ray.put(observers)
            else:
                observers_ref = observers

            if not isinstance(orbits, ObjectRef):
                orbits_ref = ray.put(orbits)
            else:
                orbits_ref = orbits
                # We need to dereference the orbits ObjectRef so we can
                # check its length for chunking and determine
                # if we need to propagate variants
                orbits = ray.get(orbits_ref)

            # Create futures
            futures_inputs = []
            idx = np.arange(0, len(orbits))
            for idx_chunk in _iterate_chunks(idx, chunk_size):
                futures_inputs.append(
                    (
                        idx_chunk,
                        orbits_ref,
                        observers_ref,
                        self,
                        predict_magnitudes,
                        predict_phase_angle,
                    )
                )

            # Add variants to propagate to futures inputs
            if covariance is True and len(variants) > 0:
                variants_ref = ray.put(variants)

                idx = np.arange(0, len(variants))
                for variant_chunk_idx in _iterate_chunks(idx, chunk_size):
                    futures_inputs.append(
                        (
                            variant_chunk_idx,
                            variants_ref,
                            observers_ref,
                            self,
                            predict_magnitudes,
                            predict_phase_angle,
                        )
                    )

            # Get results as they finish (we sort later)
            futures = []
            for future_input in futures_inputs:
                futures.append(ephemeris_worker_ray.remote(*future_input))

                if len(futures) >= max_processes * 1.5:
                    finished, futures = ray.wait(futures, num_returns=1)
                    result = ray.get(finished[0])
                    if isinstance(result, Ephemeris):
                        ephemeris = qv.concatenate([ephemeris, result])
                    elif isinstance(result, VariantEphemeris):
                        variant_ephemeris = qv.concatenate([variant_ephemeris, result])
                    else:
                        raise ValueError(
                            f"Unexpected result type from ephemeris worker: {type(result)}"
                        )

            while futures:
                finished, futures = ray.wait(futures, num_returns=1)
                result = ray.get(finished[0])
                if isinstance(result, Ephemeris):
                    ephemeris = qv.concatenate([ephemeris, result])
                elif isinstance(result, VariantEphemeris):
                    variant_ephemeris = qv.concatenate([variant_ephemeris, result])
                else:
                    raise ValueError(
                        f"Unexpected result type from ephemeris worker: {type(result)}"
                    )

            # Concatenation was in completion order; sort to canonical order.
            if len(ephemeris) > 0:
                ephemeris = ephemeris.sort_by(
                    [
                        "orbit_id",
                        "coordinates.time.days",
                        "coordinates.time.nanos",
                        "coordinates.origin.code",
                    ]
                )
            if len(variant_ephemeris) > 0:
                variant_ephemeris = variant_ephemeris.sort_by(
                    [
                        "orbit_id",
                        "variant_id",
                        "coordinates.time.days",
                        "coordinates.time.nanos",
                        "coordinates.origin.code",
                    ]
                )

        else:
            if uses_default_ephemeris:
                results = self._generate_ephemeris(
                    orbits,
                    observers,
                    predict_magnitudes=predict_magnitudes,
                    predict_phase_angle=predict_phase_angle,
                )
            else:
                results = self._generate_ephemeris(orbits, observers)
            if isinstance(results, Ephemeris):
                ephemeris = results
            elif isinstance(results, VariantEphemeris):
                variant_ephemeris = results
            else:
                raise ValueError(
                    f"Unexpected result type from generate_ephemeris: {type(results)}"
                )
            if covariance is True and len(variants) > 0:
                if uses_default_ephemeris:
                    variant_ephemeris = self._generate_ephemeris(
                        variants,
                        observers,
                        predict_magnitudes=predict_magnitudes,
                        predict_phase_angle=predict_phase_angle,
                    )
                else:
                    variant_ephemeris = self._generate_ephemeris(variants, observers)

        if covariance is False and len(variant_ephemeris) > 0:
            # If we decide that we do not need to guarantee that the time scale is in UTC

            # then we may want to call:
            # if isinstance(observers, ray.ObjectRef):
            #     variant_ephemeris = ensure_input_time_scale(
            #         variant_ephemeris, ray.get(observers).coordinates.time
            #     )
            # else:
            #     variant_ephemeris = ensure_input_time_scale(
            #         variant_ephemeris, observers.coordinates.time
            #     )
            variant_ephemeris = variant_ephemeris.set_column(
                "coordinates.time",
                variant_ephemeris.coordinates.time.rescale("utc"),
            )
            if needs_attach:
                variant_ephemeris = attach_magnitude_or_phase(
                    variant_ephemeris,
                    orbits=orbits,
                    observers=observers,
                    predict_magnitudes=predict_magnitudes,
                    predict_phase_angle=predict_phase_angle,
                )
            return variant_ephemeris

        if covariance is True and len(variant_ephemeris) > 0:
            ephemeris = variant_ephemeris.collapse(ephemeris)

        # Same note as above.
        # if isinstance(observers, ray.ObjectRef):
        #     ephemeris = ensure_input_time_scale(
        #         ephemeris, ray.get(observers).coordinates.time
        #     )
        # else:
        #     ephemeris = ensure_input_time_scale(
        #         ephemeris, observers.coordinates.time
        #     )
        ephemeris = ephemeris.set_column(
            "coordinates.time",
            ephemeris.coordinates.time.rescale("utc"),
        )
        if needs_attach:
            ephemeris = attach_magnitude_or_phase(
                ephemeris,
                orbits=orbits,
                observers=observers,
                predict_magnitudes=predict_magnitudes,
                predict_phase_angle=predict_phase_angle,
            )
        return ephemeris


class Propagator(ABC, EphemerisMixin):
    """
    Abstract class for propagating orbits and related functions.

    Subclasses should implement the _propagate_orbits.
    For additional functions, subclasses can add abstract mixins.

    Important: subclasses should be pickleable! As this class uses multiprocessing
    to parallelize propagation and ephemeris generation. This means that
    subclasses should not have any unpickleable attributes.
    """

    @abstractmethod
    def _propagate_orbits(self, orbits: OrbitType, times: TimestampType) -> OrbitType:
        """
        Propagate orbits to times.

        THIS FUNCTION SHOULD BE DEFINED BY THE USER.
        """
        pass

    def __getstate__(self):
        """
        Get the state of the propagator.

        Subclasses need to define what is picklable for multiprocessing.

        e.g.

        def __getstate__(self):
            state = self.__dict__.copy()
            state.pop("_stateful_attribute_that_is_not_pickleable")
            return state
        """
        raise NotImplementedError(
            "Propagator must implement __getstate__ for multiprocessing serialization.\n"
            "Example implementation: \n"
            "def __getstate__(self):\n"
            "    state = self.__dict__.copy()\n"
            "    state.pop('_stateful_attribute_that_is_not_pickleable')\n"
            "    return state"
        )

    def __setstate__(self, state):
        """
        Set the state of the propagator.

        Subclasses need to define what is unpicklable for multiprocessing.

        e.g.

        def __setstate__(self, state):
            self.__dict__.update(state)
            self._stateful_attribute_that_is_not_pickleable = None
        """
        raise NotImplementedError(
            "Propagator must implement __setstate__ for multiprocessing serialization.\n"
            "Example implementation: \n"
            "def __setstate__(self, state):\n"
            "    self.__dict__.update(state)\n"
            "    self._stateful_attribute_that_is_not_pickleable = None"
        )

    def propagate_orbits(
        self,
        orbits: Union[OrbitType, ObjectRef],
        times: Union[TimestampType, ObjectRef],
        covariance: bool = False,
        covariance_method: Literal[
            "auto", "sigma-point", "monte-carlo"
        ] = "monte-carlo",
        num_samples: int = 1000,
        chunk_size: int = 100,
        max_processes: Optional[int] = 1,
        seed: Optional[int] = None,
    ) -> Union[Orbits, VariantOrbits]:
        """
        Propagate each orbit in orbits to each time in times.

        This method handles parallelization of the propagation of the orbits.
        Subclasses may override this method to modify parallelization behavior.

        Parameters
        ----------
        orbits : `~adam_core.orbits.orbits.Orbits` (N)
            Orbits to propagate.
        times : Timestamp (M)
            Times to which to propagate orbits. Sorted chronologically before
            calling the backend so integrators (e.g. ASSIST, REBOUND) receive
            time-ordered epochs for efficient stepping.
        covariance : bool, optional
            Propagate the covariance matrices of the orbits. This is done by sampling the
            orbits from their covariance matrices and propagating each sample. The covariance
            of the propagated orbits is then the covariance of the samples.
        covariance_method : {'sigma-point', 'monte-carlo', 'auto'}, optional
            The method to use for sampling the covariance matrix. If 'auto' is selected then the method
            will be automatically selected based on the covariance matrix. The default is 'monte-carlo'.
        num_samples : int, optional
            The number of samples to draw when sampling with monte-carlo.
        chunk_size : int, optional
            Number of orbits to send to each job.
        max_processes : int or None, optional
            Maximum number of processes to launch. If None then the number of
            processes will be equal to the number of cores on the machine. If 1
            then no multiprocessing will be used. If "ray" is the parallel_backend and a ray instance
            is initialized already then this argument is ignored.

        Returns
        -------
        propagated : `~adam_core.orbits.orbits.Orbits` or `~adam_core.orbits.variants.VariantOrbits`
            Propagated orbits. Rows are ordered (orbit_id, time) for Orbits and
            (orbit_id, variant_id, time) for VariantOrbits. Use .sort_by() or
            .select()/.apply_mask() for other orderings or grouping.
        """
        if covariance is True and isinstance(orbits, VariantOrbits):
            raise AssertionError("Covariance is not supported for VariantOrbits")

        if max_processes is None:
            max_processes = mp.cpu_count()

        # Resolve times and sort chronologically so backends (ASSIST, REBOUND, etc.)
        # receive time-ordered epochs for efficient integration.
        if isinstance(times, ObjectRef):
            times = ray.get(times)
        times = times.sort_by(["days", "nanos"])

        if max_processes > 1:
            propagated_list: List[Orbits] = []
            covariance_variants_list: List[VariantOrbits] = []
            # When the input is VariantOrbits, do not treat them as covariance.
            propagated_variants_input_list: List[VariantOrbits] = []
            input_is_variants: Optional[bool] = None

            initialize_use_ray(num_cpus=max_processes)

            times_ref = ray.put(times)

            if not isinstance(orbits, ObjectRef):
                input_is_variants = isinstance(orbits, VariantOrbits)
                orbits_ref = ray.put(orbits)
            else:
                orbits_ref = orbits
                # We need to dereference the orbits ObjectRef so we can
                # check its length for chunking and determine
                # if we need to propagate variants
                orbits = ray.get(orbits_ref)
                input_is_variants = isinstance(orbits, VariantOrbits)

            if covariance is True and input_is_variants:
                raise AssertionError("Covariance is not supported for VariantOrbits")

            # Create futures inputs
            futures_inputs = []
            idx = np.arange(0, len(orbits))
            for idx_chunk in _iterate_chunks(idx, chunk_size):
                futures_inputs.append(
                    (
                        idx_chunk,
                        orbits_ref,
                        times_ref,
                        self,
                    )
                )

            # Add variants to propagate to futures inputs
            if covariance is True and not orbits.coordinates.covariance.is_all_nan():
                variants = VariantOrbits.create(
                    orbits,
                    method=covariance_method,
                    num_samples=num_samples,
                    seed=seed,
                )

                variants_ref = ray.put(variants)

                idx = np.arange(0, len(variants))
                for variant_chunk_idx in _iterate_chunks(idx, chunk_size):
                    futures_inputs.append(
                        (
                            variant_chunk_idx,
                            variants_ref,
                            times_ref,
                            self,
                        )
                    )

            # Submit and process jobs with queuing
            futures = []
            for future_input in futures_inputs:
                futures.append(propagation_worker_ray.remote(*future_input))

                if len(futures) >= max_processes * 1.5:
                    finished, futures = ray.wait(futures, num_returns=1)
                    result = ray.get(finished[0])
                    if isinstance(result, Orbits):
                        propagated_list.append(result)
                    elif isinstance(result, VariantOrbits):
                        if input_is_variants:
                            propagated_variants_input_list.append(result)
                        else:
                            covariance_variants_list.append(result)
                    else:
                        raise ValueError(
                            f"Unexpected result type from propagation worker: {type(result)}"
                        )

            # Process remaining futures
            while futures:
                finished, futures = ray.wait(futures, num_returns=1)
                result = ray.get(finished[0])
                if isinstance(result, Orbits):
                    propagated_list.append(result)
                elif isinstance(result, VariantOrbits):
                    if input_is_variants:
                        propagated_variants_input_list.append(result)
                    else:
                        covariance_variants_list.append(result)
                else:
                    raise ValueError(
                        f"Unexpected result type from propagation worker: {type(result)}"
                    )

            # Concatenate propagated orbits
            if input_is_variants:
                propagated = qv.concatenate(propagated_variants_input_list)
                propagated_variants = None
            else:
                propagated = qv.concatenate(propagated_list)
                if len(covariance_variants_list) > 0:
                    propagated_variants = qv.concatenate(covariance_variants_list)
                    propagated_variants = propagated_variants.sort_by(
                        [
                            "orbit_id",
                            "variant_id",
                            "coordinates.time.days",
                            "coordinates.time.nanos",
                        ]
                    )
                else:
                    propagated_variants = None

        else:
            propagated = self._propagate_orbits(orbits, times)

            if covariance is True and not orbits.coordinates.covariance.is_all_nan():
                variants = VariantOrbits.create(
                    orbits,
                    method=covariance_method,
                    num_samples=num_samples,
                    seed=seed,
                )

                propagated_variants = self._propagate_orbits(variants, times)
                propagated_variants = propagated_variants.sort_by(
                    [
                        "orbit_id",
                        "variant_id",
                        "coordinates.time.days",
                        "coordinates.time.nanos",
                    ]
                )
            else:
                propagated_variants = None

        if propagated_variants is not None:
            propagated = propagated_variants.collapse(propagated)

        # Preserve the time scale of the requested times
        propagated = ensure_input_time_scale(propagated, times)

        # Return the results with the original origin and frame
        # Preserve the original output origin for the input orbits
        # by orbit id
        propagated = ensure_input_origin_and_frame(orbits, propagated)

        # Internal order is (orbit_id, variant_id, time) so that observer tiling
        # [observers]*len(orbits) in _generate_ephemeris matches row-for-row.
        # Callers that need time-first or grouping should sort or use .select()/.apply_mask().
        if isinstance(propagated, VariantOrbits):
            return propagated.sort_by(
                [
                    "orbit_id",
                    "variant_id",
                    "coordinates.time.days",
                    "coordinates.time.nanos",
                ]
            )

        return propagated.sort_by(
            ["orbit_id", "coordinates.time.days", "coordinates.time.nanos"]
        )
